"""Plot camera pixel layout."""

import logging

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection

import simtools.visualization.legend_handlers as leg_h
from simtools.model.model_utils import is_two_mirror_telescope
from simtools.utils import names

logger = logging.getLogger(__name__)


def plot_pixel_layout(camera, camera_in_sky_coor=False, pixels_id_to_print=50):
    """
    Plot the pixel layout for an observer facing the camera.

    Including in the plot edge pixels, off pixels, pixel ID for the first 50 pixels,
    coordinate systems, FOV, focal length and the average edge radius.

    Parameters
    ----------
    camera : Camera
        Camera object to plot.
    camera_in_sky_coor : bool, optional
        Flag to plot the camera in the sky coordinate system.
    pixels_id_to_print : int, optional
        Number of pixel IDs to print in the plot.

    Returns
    -------
    fig : plt.Figure
        Figure with the pixel layout.
    """
    logger.info(f"Plotting the {camera.telescope_model_name} camera")

    fig, ax = plt.subplots(figsize=(8, 8))

    if not is_two_mirror_telescope(camera.telescope_model_name) and not camera_in_sky_coor:
        camera.pixels["y"] = [(-1) * y_val for y_val in camera.pixels["y"]]

    on_pixels, edge_pixels, off_pixels = _pixel_type_lists(camera)
    for i_pix, (x, y) in enumerate(zip(camera.pixels["x"], camera.pixels["y"])):
        if camera.pixels["pix_id"][i_pix] < pixels_id_to_print + 1:
            font_size = (
                4
                if "SCT" in names.get_array_element_type_from_name(camera.telescope_model_name)
                else 2
            )
            plt.text(
                x, y, camera.pixels["pix_id"][i_pix], ha="center", va="center", fontsize=font_size
            )
    ax.add_collection(
        PatchCollection(on_pixels, facecolor="none", edgecolor="black", linewidth=0.2)
    )
    ax.add_collection(
        PatchCollection(
            edge_pixels,
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
            linewidth=0.2,
        )
    )
    ax.add_collection(
        PatchCollection(off_pixels, facecolor="black", edgecolor="black", linewidth=0.2)
    )
    legend_objects = [leg_h.PixelObject(), leg_h.EdgePixelObject()]
    legend_labels = ["Pixel", "Edge pixel"]
    legend_handler_map = {
        leg_h.PixelObject: (
            leg_h.HexPixelHandler()
            if isinstance(on_pixels[0], mpatches.RegularPolygon)
            else leg_h.SquarePixelHandler()
        ),
        leg_h.EdgePixelObject: (
            leg_h.HexEdgePixelHandler()
            if isinstance(on_pixels[0], mpatches.RegularPolygon)
            else leg_h.SquareEdgePixelHandler()
        ),
        leg_h.OffPixelObject: (
            leg_h.HexOffPixelHandler()
            if isinstance(on_pixels[0], mpatches.RegularPolygon)
            else leg_h.SquareOffPixelHandler()
        ),
    }

    if len(off_pixels) > 0:
        legend_objects.append(leg_h.OffPixelObject())
        legend_labels.append("Disabled pixel")

    plt.axis("equal")
    plt.grid(True)
    ax.set_axisbelow(True)
    plt.axis(
        [
            min(camera.pixels["x"]),
            max(camera.pixels["x"]),
            min(camera.pixels["y"]) * 1.42,
            max(camera.pixels["y"]) * 1.42,
        ]
    )
    plt.xlabel("Horizontal scale [cm]", fontsize=18, labelpad=0)
    plt.ylabel("Vertical scale [cm]", fontsize=18, labelpad=0)
    ax.set_title(
        f"Pixels layout in {camera.telescope_model_name:s} camera",
        fontsize=15,
        y=1.02,
    )
    plt.tick_params(axis="both", which="major", labelsize=15)

    _plot_axes_def(camera, plt, camera.pixels["rotate_angle"])

    description = {
        False: "For an observer facing the camera",
        True: "For an observer behind the camera looking through",
        None: "For an observer looking from secondary to camera",
    }[camera_in_sky_coor and not is_two_mirror_telescope(camera.telescope_model_name)]
    ax.text(0.02, 0.02, description, transform=ax.transAxes, color="black", fontsize=12)

    fov, r_edge_avg = camera.calc_fov()
    ax.text(
        0.02,
        0.96,
        rf"$f_{{\mathrm{{eff}}}} = {camera.focal_length:.3f}\,\mathrm{{cm}}$",
        transform=ax.transAxes,
        color="black",
        fontsize=12,
    )
    ax.text(
        0.02,
        0.92,
        rf"Avg. edge radius = {r_edge_avg:.3f}$\,\mathrm{{cm}}$",
        transform=ax.transAxes,
        color="black",
        fontsize=12,
    )
    ax.text(
        0.02,
        0.88,
        rf"FoV = {fov:.3f}$\,\mathrm{{deg}}$",
        transform=ax.transAxes,
        color="black",
        fontsize=12,
    )
    plt.legend(
        legend_objects,
        legend_labels,
        handler_map=legend_handler_map,
        prop={"size": 11},
        loc="upper right",
    )
    ax.set_aspect("equal", "datalim")
    plt.tight_layout()

    return fig


def _pixel_type_lists(camera):
    """
    Return on, off, and edge pixel lists.

    Parameters
    ----------
    camera : Camera
        Camera object.

    Returns
    -------
    on_pixels : list
        List of on pixels.
    edge_pixels : list
        List of edge pixels.
    off_pixels : list
        List of off pixels.
    """
    on_pixels, edge_pixels, off_pixels = [], [], []

    for i_pix, (x, y) in enumerate(zip(camera.pixels["x"], camera.pixels["y"])):
        shape = _pixel_shape(camera, x, y)
        if camera.pixels["pix_on"][i_pix]:
            neighbors = camera.get_neighbor_pixels()[i_pix]
            if len(neighbors) < 6 and camera.pixels["pixel_shape"] in (1, 3):
                edge_pixels.append(shape)
            elif len(neighbors) < 4 and camera.pixels["pixel_shape"] == 2:
                edge_pixels.append(shape)
            else:
                on_pixels.append(shape)
        else:
            off_pixels.append(shape)

    return on_pixels, edge_pixels, off_pixels


def _pixel_shape(camera, x, y):
    """
    Return the shape of the pixel.

    Parameters
    ----------
    camera : Camera
        Camera object.
    x : float
        x-coordinate of the pixel.
    y : float
        y-coordinate of the pixel.

    Returns
    -------
    shape : matplotlib.patches.Patch
        Shape of the pixel.
    """
    if camera.pixels["pixel_shape"] in (1, 3):
        return mpatches.RegularPolygon(
            (x, y),
            numVertices=6,
            radius=camera.pixels["pixel_diameter"] / np.sqrt(3),
            orientation=np.deg2rad(camera.pixels["orientation"]),
        )
    if camera.pixels["pixel_shape"] == 2:
        return mpatches.Rectangle(
            (
                x - camera.pixels["pixel_diameter"] / 2.0,
                y - camera.pixels["pixel_diameter"] / 2.0,
            ),
            width=camera.pixels["pixel_diameter"],
            height=camera.pixels["pixel_diameter"],
        )
    return None


def _plot_axes_def(camera, plot, rotate_angle):
    """
    Plot three axes definitions on the pyplot.plt instance provided.

    The three axes are Alt/Az, the camera coordinate system and the original coordinate
    system the pixel list was provided.

    Parameters
    ----------
    camera : Camera
        Camera object.
    plot: pyplot.plt instance
        A pyplot.plt instance where to add the axes definitions.
    rotate_angle: float
        The rotation angle applied
    """
    invert_yaxis = False
    x_left = 0.7  # Position of the left most axis
    if not is_two_mirror_telescope(camera.telescope_model_name):
        invert_yaxis = True
        x_left = 0.8

    x_title = r"$x_{\!pix}$"
    y_title = r"$y_{\!pix}$"
    x_pos, y_pos = (x_left, 0.12)
    # The rotation of LST (above 100 degrees) raises the axes.
    # In this case, lower the starting point.
    if np.rad2deg(rotate_angle) > 100:
        y_pos -= 0.09
        x_pos -= 0.05
    kwargs = {
        "x_title": x_title,
        "y_title": y_title,
        "x_pos": x_pos,
        "y_pos": y_pos,
        "rotate_angle": rotate_angle - (1 / 2.0) * np.pi,
        "fc": "black",
        "ec": "black",
        "invert_yaxis": invert_yaxis,
    }
    _plot_one_axis_def(plot, **kwargs)

    x_title = r"$x_{\!cam}$"
    y_title = r"$y_{\!cam}$"
    x_pos, y_pos = (x_left + 0.15, 0.12)
    kwargs = {
        "x_title": x_title,
        "y_title": y_title,
        "x_pos": x_pos,
        "y_pos": y_pos,
        "rotate_angle": (3 / 2.0) * np.pi,
        "fc": "blue",
        "ec": "blue",
        "invert_yaxis": invert_yaxis,
    }
    _plot_one_axis_def(plot, **kwargs)

    x_title = "Alt"
    y_title = "Az"
    x_pos, y_pos = (x_left + 0.15, 0.25)
    kwargs = {
        "x_title": x_title,
        "y_title": y_title,
        "x_pos": x_pos,
        "y_pos": y_pos,
        "rotate_angle": (3 / 2.0) * np.pi,
        "fc": "red",
        "ec": "red",
        "invert_yaxis": invert_yaxis,
    }
    _plot_one_axis_def(plot, **kwargs)


def _plot_one_axis_def(plot, **kwargs):
    """
    Plot an axis on the pyplot.plt instance provided.

    Parameters
    ----------
    plot: pyplot.plt instance
        A pyplot.plt instance where to add the axes definitions.
    **kwargs: dict
            x_title: str
            x-axis title
            y_title: str
            y-axis title,
            x_pos: float
            x position of the axis to draw
            y_pos: float
            y position of the axis to draw
            rotate_angle: float
            rotation angle of the axis in radians
            fc: str
            face colour of the axis
            ec: str
            edge colour of the axis
            invert_yaxis: bool
            Flag to invert the y-axis (for dual mirror telescopes).
    """
    x_title = kwargs["x_title"]
    y_title = kwargs["y_title"]
    x_pos, y_pos = (kwargs["x_pos"], kwargs["y_pos"])

    r = 0.1  # size of arrow
    sign = 1.0
    if kwargs["invert_yaxis"]:
        sign *= -1.0
    x_text1 = x_pos + sign * r * np.cos(kwargs["rotate_angle"])
    y_text1 = y_pos + r * np.sin(0 + kwargs["rotate_angle"])
    x_text2 = x_pos + sign * r * np.cos(np.pi / 2.0 + kwargs["rotate_angle"])
    y_text2 = y_pos + r * np.sin(np.pi / 2.0 + kwargs["rotate_angle"])

    plot.gca().annotate(
        x_title,
        xy=(x_pos, y_pos),
        xytext=(x_text1, y_text1),
        xycoords="axes fraction",
        ha="center",
        va="center",
        size="xx-large",
        arrowprops={
            "arrowstyle": "<|-",
            "shrinkA": 0,
            "shrinkB": 0,
            "fc": kwargs["fc"],
            "ec": kwargs["ec"],
        },
    )

    plot.gca().annotate(
        y_title,
        xy=(x_pos, y_pos),
        xytext=(x_text2, y_text2),
        xycoords="axes fraction",
        ha="center",
        va="center",
        size="xx-large",
        arrowprops={
            "arrowstyle": "<|-",
            "shrinkA": 0,
            "shrinkB": 0,
            "fc": kwargs["fc"],
            "ec": kwargs["ec"],
        },
    )



#!/usr/bin/env python
r"""
Calculate photon incident angles on focal plane and primary/secondary mirrors.

Creates photon files with additional columns for incident angles calculation.
Outputs files and histograms of the incidence angles at
the focal plane, primary mirror, and if available, secondary mirror.
Optional debug plots can be also generated.
Note that this application does not include a full raytracing of telescope structures,
and their non symmetric shadowing at off-axis angles.

Example usage
-------------

.. code-block:: console

    simtools-calculate-incident-angles \
        --off_axis_angles 0 1 2 3 4 \
        --source_distance 10 \
        --number_of_photons 1000000 \
        --model_version 6.0.0 \
        --telescope MSTN-04 \
        --site North

Command line arguments
----------------------

- off_axis_angles (float, optional)
    One or more off-axis angles in degrees (space-separated). Default: [0.0].
- source_distance (float, optional)
    Source distance in kilometers. Default: 10.0.
- number_of_photons (int, optional)
    Number of photons of the light source to trace per run. Default: 10000.
- perfect_mirror (flag, optional)
    Assume perfect mirror shape/alignment/reflection.
- debug_plots (flag, optional)
    Generate additional debug plots (radius histograms, XY heatmaps, radius vs angle).
- calculate_primary_secondary_angles / no-calculate_primary_secondary_angles
    Include or skip angles on primary/secondary mirrors. Default: include.

The application writes:

- imaging list (photons) file
- stars list file
- a histogram of incident angles (PNG)
- a results table in ECSV format

Example of a focal-plane incident angle plot for a SST:

.. _plot_calculate_incident_angles_plot:
.. image:: images/incident_angles_multi_calculate_incident_angles_SSTS-04.png
    :width: 49 %

Example of a primary mirror incident angle plot for a SST:

.. _plot_calculate_incident_angles_plot_primary:
.. image:: images/incident_angles_primary_multi_calculate_incident_angles_SSTS-04.png
    :width: 49 %

Note also the relation between radius and primary mirror incident angles, and how this relates to
the peak seen in the primary mirror incident angle distribution:

.. _plot_calculate_incident_angles_plot_angle_vs_radius:
.. image:: images/primary_angle_vs_radius.png
    :width: 49 %

Example of a secondary mirror incident angle plot for a SST:

.. _plot_calculate_incident_angles_plot_secondary:
.. image:: images/incident_angles_secondary_multi_calculate_incident_angles_SSTS-04.png
    :width: 49 %
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.configuration.commandline_parser import CommandLineParser
from simtools.ray_tracing.incident_angles import IncidentAnglesCalculator
from simtools.visualization.plot_incident_angles import plot_incident_angles


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Calculate photon incident angles on focal plane and primary/secondary mirrors."
        ),
    )
    config.parser.add_argument(
        "--off_axis_angles",
        help="One or more off-axis angles in degrees (space-separated)",
        type=float,
        nargs="+",
        required=False,
    )
    config.parser.add_argument(
        "--source_distance",
        help="Source distance in kilometers",
        type=float,
        default=10.0,
        required=False,
    )
    config.parser.add_argument(
        "--number_of_photons",
        help="Number of star photons to trace (per run)",
        type=CommandLineParser.scientific_int,
        default=10000,
        required=False,
    )
    config.parser.add_argument(
        "--perfect_mirror",
        help="Assume perfect mirror shape/alignment/reflection",
        action="store_true",
        required=False,
    )
    config.parser.add_argument(
        "--debug_plots",
        dest="debug_plots",
        help="Generate additional debug plots (radius histograms, XY heatmaps, radius vs angle)",
        action="store_true",
        required=False,
    )
    config.parser.add_argument(
        "--calculate_primary_secondary_angles",
        dest="calculate_primary_secondary_angles",
        help="Also compute angles of incidence on primary and secondary mirrors",
        required=False,
        action="store_true",
    )
    return config.initialize(
        db_config=True,
        simulation_model=["telescope", "site", "model_version"],
    )


def main():
    """Calculate photon incident angles on focal plane and primary/secondary mirrors."""
    app_context = startup_application(_parse)

    app_context.logger.info("Starting calculation of incident angles")

    output_dir = app_context.io_handler.get_output_directory()
    base_label = app_context.args.get("label", get_application_label(__file__))
    telescope_name = app_context.args["telescope"]
    label_with_telescope = f"{base_label}_{telescope_name}"

    calculator = IncidentAnglesCalculator(
        simtel_path=app_context.args["simtel_path"],
        db_config=app_context.db_config,
        config_data=app_context.args,
        output_dir=output_dir,
        label=base_label,
    )
    offsets = [float(v) for v in app_context.args.get("off_axis_angles", [0.0])]

    results_by_offset = calculator.run_for_offsets(offsets)
    plot_incident_angles(
        results_by_offset,
        output_dir,
        label_with_telescope,
        debug_plots=app_context.args.get("debug_plots", False),
    )
    total = sum(len(t) for t in results_by_offset.values())
    summary_msg = (
        f"Calculated incident angles for {len(results_by_offset)} offsets,\n"
        f"total photon statistics {total}"
    )
    if total < 1_000_000:
        summary_msg += " (below 1e6; results may be statistically unstable)"
    app_context.logger.info(summary_msg)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
    Calculate on-axis telescope throughput and NSB pixels rates.

    Uses the sim_telarray tool "testeff" to calculate the camera efficiency.
    The results of telescope throughput including optical and camera components for Cherenkov (left)
    and NSB light (right) as a function of wavelength are plotted. See examples below.

    .. _validate_camera_eff_plot:
    .. image:: images/validate_camera_efficiency_North-MST-NectarCam-D_cherenkov.png
      :width: 49 %
    .. image:: images/validate_camera_efficiency_North-MST-NectarCam-D_nsb.png
      :width: 49 %

    Command line arguments
    ----------------------
    site (str, required)
        North or South.
    telescope (str, required)
        Telescope model name (e.g. LSTN-01, SSTS-15)
    model_version (str, optional)
        Simulation model version
    zenith_angle (float, optional)
        Zenith angle in degrees (between 0 and 180).
    azimuth_angle (float, optional)
        Telescope pointing direction in azimuth.
    nsb_spectrum (str, optional)
        File with NSB spectrum to use for the efficiency simulation.

    Example
    -------
    MSTN-01 5.0.0

    Runtime < 1 min.

    .. code-block:: console

        simtools-validate-camera-efficiency --site North \\
            --azimuth_angle 0 --zenith_angle 20 \\
            --nsb_spectrum average_nsb_spectrum_CTAO-N_ze20_az0.txt \\
            --telescope MSTN-01 --model_version 5.0.0

    The output is saved in simtools-output/validate_camera_efficiency.
"""

from pathlib import Path

import simtools.data_model.model_data_writer as writer
from simtools.application_control import get_application_label, startup_application
from simtools.camera.camera_efficiency import CameraEfficiency
from simtools.configuration import configurator


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Calculate the camera efficiency and NSB pixel rates. "
            "Plot the camera efficiency vs wavelength for Cherenkov and NSB light."
        ),
    )
    config.parser.add_argument(
        "--nsb_spectrum",
        help=(
            "File with NSB spectrum to use for the efficiency simulation."
            "The expected format is two columns with wavelength in nm and "
            "NSB flux with the units: [1e9 * ph/m2/s/sr/nm]."
            "If the file has more than two columns, the first and third are used,"
            "and the second is ignored (native sim_telarray behavior)."
        ),
        type=str,
        default=None,
        required=False,
    )
    config.parser.add_argument(
        "--skip_correction_to_nsb_spectrum",
        help=(
            "Skip correction to the NSB spectrum to account for the "
            "difference between the altitude used in the reference B&E spectrum and "
            "the observation level at the CTAO sites."
        ),
        required=False,
        action="store_true",
    )
    config.parser.add_argument(
        "--write_reference_nsb_rate_as_parameter",
        help=("Write the NSB pixel rate obtained for reference conditions as a model parameter "),
        action="store_true",
        required=False,
    )
    args_dict, db_config = config.initialize(
        db_config=True,
        simulation_model=["telescope", "model_version", "parameter_version"],
        simulation_configuration={"corsika_configuration": ["zenith_angle", "azimuth_angle"]},
    )
    if args_dict["site"] is None or args_dict["telescope"] is None:
        config.parser.print_help()
        print("\n\nSite and telescope must be provided\n\n")
        raise RuntimeError("Site and telescope must be provided")
    return args_dict, db_config


def main():
    """Calculate the camera efficiency and NSB pixel rates."""
    app_context = startup_application(_parse)

    ce = CameraEfficiency(
        db_config=app_context.db_config,
        label=app_context.args.get("label"),
        config_data=app_context.args,
    )
    ce.simulate()
    ce.analyze(force=True)
    ce.plot_efficiency(efficiency_type="Cherenkov", save_fig=True)
    ce.plot_efficiency(efficiency_type="NSB", save_fig=True)

    writer.ModelDataWriter.dump_model_parameter(
        parameter_name="nsb_pixel_rate",
        value=ce.get_nsb_pixel_rate(
            reference_conditions=app_context.args.get(
                "write_reference_nsb_rate_as_parameter", False
            )
        ),
        instrument=app_context.args["telescope"],
        parameter_version=app_context.args.get("parameter_version", "0.0.0"),
        output_file=Path(
            f"nsb_pixel_rate-{app_context.args.get('parameter_version', '0.0.0')}.json"
        ),
        output_path=app_context.io_handler.get_output_directory()
        / app_context.args["telescope"]
        / "nsb_pixel_rate",
    )


if __name__ == "__main__":
    main()



"""Generate a reduced dataset from sim_telarray output files using astropy tables."""

import logging
from dataclasses import dataclass

import astropy.units as u
import numpy as np
from astropy.table import Table
from eventio import EventIOFile
from eventio.simtel import (
    ArrayEvent,
    MCEvent,
    MCRunHeader,
    MCShower,
    TrackingPosition,
    TriggerInformation,
)

from simtools.corsika.primary_particle import PrimaryParticle
from simtools.simtel.simtel_io_file_info import get_corsika_run_header
from simtools.simtel.simtel_io_metadata import (
    get_sim_telarray_telescope_id_to_telescope_name_mapping,
    read_sim_telarray_metadata,
)
from simtools.utils.geometry import calculate_circular_mean
from simtools.utils.names import get_common_identifier_from_array_element_name


@dataclass
class TableSchemas:
    """Define schemas for output tables with units."""

    shower_schema = {
        "shower_id": (np.uint32, None),
        "event_id": (np.uint32, None),
        "file_id": (np.uint32, None),
        "simulated_energy": (np.float64, u.TeV),
        "x_core": (np.float64, u.m),
        "y_core": (np.float64, u.m),
        "shower_azimuth": (np.float64, u.deg),
        "shower_altitude": (np.float64, u.deg),
        "area_weight": (np.float64, None),
    }

    trigger_schema = {
        "shower_id": (np.uint32, None),
        "event_id": (np.uint32, None),
        "file_id": (np.uint32, None),
        "array_altitude": (np.float64, u.deg),
        "array_azimuth": (np.float64, u.deg),
        "telescope_list": (str, None),  # Store as comma-separated string
        "telescope_list_common_id": (str, None),  # Store as comma-separated string
    }

    file_info_schema = {
        "file_name": (str, None),
        "file_id": (np.uint32, None),
        "particle_id": (np.uint32, None),
        "energy_min": (np.float64, u.TeV),
        "energy_max": (np.float64, u.TeV),
        "viewcone_min": (np.float64, u.deg),
        "viewcone_max": (np.float64, u.deg),
        "core_scatter_min": (np.float64, u.m),
        "core_scatter_max": (np.float64, u.m),
        "zenith": (np.float64, u.deg),
        "azimuth": (np.float64, u.deg),
        "nsb_level": (np.float64, None),
    }


class SimtelIOEventDataWriter:
    """
    Process sim_telarray events and write tables to file.

    Extracts essential information from sim_telarray output files:

    - Shower parameters (energy, core location, direction)
    - Trigger patterns
    - Telescope pointing

    Attributes
    ----------
    input_files : list
        List of input file paths to process.
    max_files : int, optional
        Maximum number of files to process.
    """

    def __init__(self, input_files, max_files=100):
        """Initialize class."""
        self._logger = logging.getLogger(__name__)
        self.input_files = input_files
        try:
            self.max_files = max_files if max_files < len(input_files) else len(input_files)
        except TypeError as exc:
            raise TypeError("No input files provided.") from exc

        self.n_use = None
        self.shower_data = []
        self.trigger_data = []
        self.file_info = []
        self.telescope_id_to_name = {}

    def process_files(self):
        """
        Process input files and return tables.

        Returns
        -------
        list
            List of astropy tables containing processed data.
        """
        for i, file in enumerate(self.input_files[: self.max_files]):
            self._logger.info(f"Processing file {i + 1}/{self.max_files}: {file}")
            self._process_file(i, file)

        return self.create_tables()

    def create_tables(self):
        """Create astropy tables from collected data."""
        tables = []
        for data, schema, name in [
            (self.shower_data, TableSchemas.shower_schema, "SHOWERS"),
            (self.trigger_data, TableSchemas.trigger_schema, "TRIGGERS"),
            (self.file_info, TableSchemas.file_info_schema, "FILE_INFO"),
        ]:
            table = Table(rows=data, names=schema.keys())
            table.meta["EXTNAME"] = name
            self._add_units_to_table(table, schema)
            tables.append(table)
        return tables

    def _add_units_to_table(self, table, schema):
        """Add units to a single table's columns."""
        for col, (_, unit) in schema.items():
            if unit is not None:
                table[col].unit = unit

    def _process_file(self, file_id, file):
        """Process a single file and update data lists."""
        self._process_file_info(file_id, file)
        with EventIOFile(file) as f:
            for eventio_object in f:
                if isinstance(eventio_object, MCRunHeader):
                    self._process_mc_run_header(eventio_object)
                elif isinstance(eventio_object, MCShower):
                    self._process_mc_shower(eventio_object, file_id)
                elif isinstance(eventio_object, MCEvent):
                    self._process_mc_event(eventio_object)
                elif isinstance(eventio_object, ArrayEvent):
                    self._process_array_event(eventio_object, file_id)

    def _process_mc_run_header(self, eventio_object):
        """Process MC run header and update data lists."""
        mc_head = eventio_object.parse()
        self.n_use = mc_head["n_use"]  # reuse factor n_use needed to extend the values below
        self._logger.info(f"Shower reuse factor: {self.n_use} (viewcone: {mc_head['viewcone']})")

    def _process_file_info(self, file_id, file):
        """Process file information and append to file info list."""
        run_info = get_corsika_run_header(file)
        self.telescope_id_to_name = get_sim_telarray_telescope_id_to_telescope_name_mapping(file)
        particle = PrimaryParticle(
            particle_id_type="eventio_id", particle_id=run_info.get("primary_id", 1)
        )
        self.file_info.append(
            {
                "file_name": str(file),
                "file_id": file_id,
                "particle_id": particle.corsika7_id,
                "energy_min": run_info["E_range"][0],
                "energy_max": run_info["E_range"][1],
                "viewcone_min": run_info["viewcone"][0],
                "viewcone_max": run_info["viewcone"][1],
                "core_scatter_min": run_info["core_range"][0],
                "core_scatter_max": run_info["core_range"][1],
                "zenith": 90.0 - np.degrees(run_info["direction"][1]),
                "azimuth": np.degrees(run_info["direction"][0]),
                "nsb_level": self.get_nsb_level_from_sim_telarray_metadata(file),
            }
        )

    def _process_mc_shower(self, eventio_object, file_id):
        """
        Process MC shower and update shower event list.

        Duplicated entries 'self.n_use' times to match the number simulated events with
        different core positions.
        """
        shower = eventio_object.parse()

        self.shower_data.extend(
            {
                "shower_id": shower["shower"],
                "event_id": None,  # filled in _process_mc_event
                "file_id": file_id,
                "simulated_energy": shower["energy"],
                "x_core": None,  # filled in _process_mc_event
                "y_core": None,  # filled in _process_mc_event
                "shower_azimuth": np.degrees(shower["azimuth"]),
                "shower_altitude": np.degrees(shower["altitude"]),
                "area_weight": None,  # filled in _process_mc_event
            }
            for _ in range(self.n_use)
        )

    def _process_mc_event(self, eventio_object):
        """
        Process MC event and update shower event list.

        Expected to be called n_use times after _process_shower.
        """
        event = eventio_object.parse()

        shower_data_index = len(self.shower_data) - self.n_use + event["event_id"] % 100

        try:
            if self.shower_data[shower_data_index]["shower_id"] != event["shower_num"]:
                raise IndexError
        except IndexError as exc:
            raise IndexError(
                f"Inconsistent shower and MC event data for shower id {event['shower_num']}"
            ) from exc

        self.shower_data[shower_data_index].update(
            {
                "event_id": event["event_id"],
                "x_core": event["xcore"],
                "y_core": event["ycore"],
                "area_weight": event["aweight"],
            }
        )

    def _process_array_event(self, eventio_object, file_id):
        """Process array event and update triggered event list."""
        tracking_positions = []
        telescopes = []

        for obj in eventio_object:
            if isinstance(obj, TriggerInformation):
                trigger_info = obj.parse()
                telescopes = (
                    trigger_info["triggered_telescopes"]
                    if len(trigger_info["triggered_telescopes"]) > 0
                    else []
                )
            if isinstance(obj, TrackingPosition):
                tracking_position = obj.parse()
                tracking_positions.append(
                    {
                        "altitude": np.degrees(tracking_position["altitude_raw"]),
                        "azimuth": np.degrees(tracking_position["azimuth_raw"]),
                    }
                )

        if len(telescopes) > 0 and tracking_positions:
            self._fill_array_event(
                self._map_telescope_names(telescopes),
                tracking_positions,
                eventio_object.event_id,
                file_id,
            )

    def _fill_array_event(self, telescopes, tracking_positions, event_id, file_id):
        """Add array event triggered events with tracking positions."""
        altitudes = [pos["altitude"] for pos in tracking_positions]
        azimuths = [pos["azimuth"] for pos in tracking_positions]

        self.trigger_data.append(
            {
                "shower_id": self.shower_data[-1]["shower_id"],
                "event_id": event_id,
                "file_id": file_id,
                "array_altitude": float(np.mean(altitudes)),
                "array_azimuth": float(np.degrees(calculate_circular_mean(np.deg2rad(azimuths)))),
                "telescope_list": ",".join(map(str, telescopes)),
                "telescope_list_common_id": ",".join(
                    [
                        str(get_common_identifier_from_array_element_name(tel, 0))
                        for tel in telescopes
                    ]
                ),
            }
        )

    def _map_telescope_names(self, telescope_ids):
        """
        Map sim_telarray telescopes IDs to CTAO array element names.

        Parameters
        ----------
        telescope_ids : list
            List of telescope IDs.

        Returns
        -------
        list
            List of telescope names corresponding to the IDs.
        """
        return [
            self.telescope_id_to_name.get(tel_id, f"Unknown_{tel_id}") for tel_id in telescope_ids
        ]

    def get_nsb_level_from_sim_telarray_metadata(self, file):
        """
        Return NSB level from sim_telarray metadata.

        Falls back to preliminary NSB level if not found.

        Parameters
        ----------
        file : Path
            Path to the sim_telarray file.

        Returns
        -------
        float
            NSB level.
        """
        metadata, _ = read_sim_telarray_metadata(file)
        nsb_integrated_flux = metadata.get("nsb_integrated_flux")
        return nsb_integrated_flux or self._get_nsb_level_from_file_name(str(file))

    def _get_nsb_level_from_file_name(self, file):
        """
        Return NSB level from file name.

        Hardwired values are used for "dark", "half", and "full" NSB levels.
        Allows to read legacy sim_telarray files without 'nsb_integrated_flux'
        metadata field.

        Parameters
        ----------
        file : str
            File name to extract NSB level from.

        Returns
        -------
        float
            NSB level extracted from file name.
        """
        nsb_levels = {"dark": 0.24, "half": 0.835, "full": 1.2}

        for key, value in nsb_levels.items():
            try:
                if key in file.lower():
                    self._logger.warning(f"NSB level set to hardwired value of {value} for {file}")
                    return value
            except AttributeError as exc:
                raise AttributeError("Invalid file name.") from exc

        self._logger.warning(f"No NSB level found in {file}, defaulting to None")
        return None



r"""
Compare two directories with model production tables in JSON format.

This script should be used to support the maintenance the simulation model repository.

Example
-------
.. code-block:: console

    simtools-maintain-simulation-model-compare-productions \\
        --directory_1 ../simulation-models-dev/simulation-models/6.0.0/ \\
        --directory_2 ../simulation-models-dev/simulation-models/6.2.0

"""

from pathlib import Path

import simtools.utils.general as gen
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.io import ascii_handler


def _parse():
    """Parse command line arguments."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Compare two directories with model production tables in JSON format.",
    )
    config.parser.add_argument(
        "--directory_1",
        type=str,
        required=True,
        help="Path to the first directory containing JSON files.",
    )
    config.parser.add_argument(
        "--directory_2",
        type=str,
        required=True,
        help="Path to the second directory containing JSON files.",
    )
    return config.initialize(db_config=False, output=False)


def _print_differences(differences, rel_path):
    """Print differences in a readable format."""
    print(f"Difference in {rel_path}:\n{'-' * 40}")
    for diff in differences:
        # Clean up the path formatting for better readability
        clean_diff = diff.replace("['parameters']['", "parameters.").replace("']['", ".")
        clean_diff = clean_diff.replace("['", "").replace("']", "")
        print(f"  {clean_diff}")
    print(f"{'-' * 40}\n")


def _compare_json_dirs(dir1, dir2, ignore_key="model_version"):
    """Compare two directories containing JSON files, ignoring a specific key."""
    for path1 in dir1.rglob("*.json"):
        rel_path = path1.relative_to(dir1)
        path2 = dir2 / rel_path

        if not path2.exists():
            print(f"Missing in dir2: {rel_path}")
            continue

        try:
            json1 = gen.remove_key_from_dict(
                ascii_handler.collect_data_from_file(path1), ignore_key
            )
            json2 = gen.remove_key_from_dict(
                ascii_handler.collect_data_from_file(path2), ignore_key
            )
        except FileNotFoundError as e:
            print(f"Error reading {rel_path}: {e}")
            continue

        differences = gen.find_differences_in_json_objects(json1, json2)
        if differences:
            _print_differences(differences, rel_path)

    # Check for files present in dir2 but not dir1
    for path2 in dir2.rglob("*.json"):
        rel_path = path2.relative_to(dir2)
        if not (dir1 / rel_path).exists():
            print(f"Missing in dir1: {rel_path}")


def main():
    """Compare two directories with model production tables in JSON format."""
    app_context = startup_application(_parse, setup_io_handler=False)

    _compare_json_dirs(Path(app_context.args["directory_1"]), Path(app_context.args["directory_2"]))


if __name__ == "__main__":
    main()



"""Derive CORSIKA limits from a reduced event data file."""

import datetime
import logging

import astropy.units as u
import numpy as np
from astropy.table import Column, Table

from simtools.data_model.metadata_collector import MetadataCollector
from simtools.io import ascii_handler, io_handler
from simtools.layout.array_layout_utils import get_array_elements_from_db_for_layouts
from simtools.simtel.simtel_io_event_histograms import SimtelIOEventHistograms
from simtools.visualization import plot_simtel_event_histograms

_logger = logging.getLogger(__name__)


def generate_corsika_limits_grid(args_dict, db_config=None):
    """
    Generate CORSIKA limits.

    Parameters
    ----------
    args_dict : dict
        Dictionary containing command line arguments.
    db_config : dict, optional
        Database configuration dictionary.
    """
    if args_dict.get("array_layout_name"):
        telescope_configs = get_array_elements_from_db_for_layouts(
            args_dict["array_layout_name"],
            args_dict.get("site"),
            args_dict.get("model_version"),
            db_config,
        )
    else:
        telescope_configs = ascii_handler.collect_data_from_file(args_dict["telescope_ids"])[
            "telescope_configs"
        ]

    results = []
    for array_name, telescope_ids in telescope_configs.items():
        _logger.info(
            f"Processing file: {args_dict['event_data_file']} with telescope config: {array_name}"
        )
        result = _process_file(
            args_dict["event_data_file"],
            array_name,
            telescope_ids,
            args_dict["loss_fraction"],
            args_dict["plot_histograms"],
        )
        result["layout"] = array_name
        results.append(result)

    write_results(results, args_dict)


def _process_file(file_path, array_name, telescope_ids, loss_fraction, plot_histograms):
    """
    Compute limits for a given event data file and telescope configuration.

    Compute limits for energy, radial distance, and viewcone.

    Parameters
    ----------
    file_path : str
        Path to the event data file.
    array_name : str
        Name of the telescope array configuration.
    telescope_ids : list[int]
        List of telescope IDs to filter the events.
    loss_fraction : float
        Fraction of events to be lost.
    plot_histograms : bool
        Whether to plot histograms.

    Returns
    -------
    dict
        Dictionary containing the computed limits and metadata.
    """
    histograms = SimtelIOEventHistograms(
        file_path, array_name=array_name, telescope_list=telescope_ids
    )
    histograms.fill()

    limits = {
        "lower_energy_limit": compute_lower_energy_limit(histograms, loss_fraction),
        "upper_radius_limit": compute_upper_radius_limit(histograms, loss_fraction),
        "viewcone_radius": compute_viewcone(histograms, loss_fraction),
    }

    if plot_histograms:
        plot_simtel_event_histograms.plot(
            histograms.histograms,
            output_path=io_handler.IOHandler().get_output_directory(),
            limits=limits,
            array_name=array_name,
        )

    return limits


def write_results(results, args_dict):
    """
    Write the computed limits as astropy table to file.

    Parameters
    ----------
    results : list[dict]
        List of computed limits.
    args_dict : dict
        Dictionary containing command line arguments.
    """
    table = _create_results_table(results, args_dict["loss_fraction"])

    output_dir = io_handler.IOHandler().get_output_directory()
    output_file = output_dir / args_dict["output_file"]

    table.write(output_file, format="ascii.ecsv", overwrite=True)
    _logger.info(f"Results saved to {output_file}")

    MetadataCollector.dump(args_dict, output_file)


def _create_results_table(results, loss_fraction):
    """
    Convert list of simulation results to an astropy Table with metadata.

    Round values to appropriate precision and add metadata.

    Parameters
    ----------
    results : list[dict]
        Computed limits per file and telescope configuration.
    loss_fraction : float
        Fraction of lost events (added to metadata).

    Returns
    -------
    astropy.table.Table
        Table with computed limits.
    """
    cols = [
        "primary_particle",
        "array_name",
        "telescope_ids",
        "zenith",
        "azimuth",
        "nsb_level",
        "lower_energy_limit",
        "upper_radius_limit",
        "viewcone_radius",
    ]

    columns = {name: [] for name in cols}
    units = {}

    for res in results:
        _process_result_row(res, cols, columns, units)

    table_cols = _create_table_columns(cols, columns, units)
    table = Table(table_cols)

    table.meta.update(
        {
            "created": datetime.datetime.now().isoformat(),
            "description": "Lookup table for CORSIKA limits computed from simulations.",
            "loss_fraction": loss_fraction,
        }
    )

    return table


def _process_result_row(res, cols, columns, units):
    """Process a single result row and add values to columns."""
    for k in cols:
        val = res.get(k, None)
        if val is not None:
            val = _round_value(k, val)
            _logger.debug(f"Adding {k}: {val} to column data")

        if hasattr(val, "unit"):
            columns[k].append(val.value)
            units[k] = val.unit
        else:
            columns[k].append(val)
            if k not in units:
                units[k] = None


def _round_value(key, val):
    """Round value based on key type."""
    if key == "lower_energy_limit":
        return np.floor(val * 1e3) / 1e3
    if key == "upper_radius_limit":
        return np.ceil(val / 25) * 25
    if key == "viewcone_radius":
        return np.ceil(val / 0.25) * 0.25
    return val


def _create_table_columns(cols, columns, units):
    """Create table columns with appropriate data types."""
    table_cols = []
    for k in cols:
        col_data = columns[k]
        if any(isinstance(v, list | tuple) for v in col_data):
            col = Column(data=col_data, name=k, unit=units.get(k), dtype=object)
        else:
            col = Column(data=col_data, name=k, unit=units.get(k))
        table_cols.append(col)
    return table_cols


def _compute_limits(hist, bin_edges, loss_fraction, limit_type="lower"):
    """
    Compute the limits based on the loss fraction.

    Add or subtract one bin to be on the safe side of the limit.

    Parameters
    ----------
    hist : np.ndarray
        1D histogram array.
    bin_edges : np.ndarray
        Array of bin edges.
    loss_fraction : float
        Fraction of events to be lost.
    limit_type : str, optional
        Type of limit ('lower' or 'upper'). Default is 'lower'.

    Returns
    -------
    float
        Bin edge value corresponding to the threshold.
    """
    total_events = np.sum(hist)
    threshold = (1 - loss_fraction) * total_events
    if limit_type == "upper":
        cum = np.cumsum(hist)
        idx = np.searchsorted(cum, threshold) + 1
        return bin_edges[min(idx, len(bin_edges) - 1)]
    if limit_type == "lower":
        cum = np.cumsum(hist[::-1])
        idx = np.searchsorted(cum, threshold) + 1
        return bin_edges[max(len(bin_edges) - 1 - idx, 0)]
    raise ValueError("limit_type must be 'lower' or 'upper'")


def compute_lower_energy_limit(histograms, loss_fraction):
    """
    Compute the lower energy limit in TeV based on the event loss fraction.

    Parameters
    ----------
    histograms : SimtelIOEventHistograms
        Histograms.
    loss_fraction : float
        Fraction of events to be lost.

    Returns
    -------
    astropy.units.Quantity
        Lower energy limit.
    """
    energy_min = (
        _compute_limits(
            histograms.histograms["energy"]["histogram"],
            histograms.energy_bins,
            loss_fraction,
            limit_type="lower",
        )
        * u.TeV
    )

    return _is_close(
        energy_min,
        histograms.file_info["energy_min"].to("TeV")
        if "energy_min" in histograms.file_info
        else None,
        "Lower energy limit is equal to the minimum energy of",
    )


def _is_close(value, reference, warning_text):
    """Check if the value is close to the reference value and log a warning if so."""
    if reference is not None and np.isclose(value.value, reference.value, rtol=1.0e-2):
        _logger.warning(f"{warning_text} {value}.")
    return value


def compute_upper_radius_limit(histograms, loss_fraction):
    """
    Compute the upper radial distance based on the event loss fraction.

    Parameters
    ----------
    histograms : SimtelIOEventHistograms
        Histograms.
    loss_fraction : float
        Fraction of events to be lost.

    Returns
    -------
    astropy.units.Quantity
        Upper radial distance in m.
    """
    radius_limit = (
        _compute_limits(
            histograms.histograms["core_distance"]["histogram"],
            histograms.core_distance_bins,
            loss_fraction,
            limit_type="upper",
        )
        * u.m
    )
    return _is_close(
        radius_limit,
        histograms.file_info["core_scatter_max"].to("m")
        if "core_scatter_max" in histograms.file_info
        else None,
        "Upper radius limit is equal to the maximum core scatter distance of",
    )


def compute_viewcone(histograms, loss_fraction):
    """
    Compute the viewcone based on the event loss fraction.

    The shower IDs of triggered events are used to create a mask for the
    azimuth and altitude of the triggered events. A mapping is created
    between the triggered events and the simulated events using the shower IDs.

    Parameters
    ----------
    histograms : SimtelIOEventHistograms
        Histograms.
    loss_fraction : float
        Fraction of events to be lost.

    Returns
    -------
    astropy.units.Quantity
        Viewcone radius in degrees.
    """
    viewcone_limit = (
        _compute_limits(
            histograms.histograms["angular_distance"]["histogram"],
            histograms.view_cone_bins,
            loss_fraction,
            limit_type="upper",
        )
        * u.deg
    )
    return _is_close(
        viewcone_limit,
        histograms.file_info["viewcone_max"].to("deg")
        if "viewcone_max" in histograms.file_info
        else None,
        "Upper viewcone limit is equal to the maximum viewcone distance of",
    )



"""Definition and modeling of camera."""

import logging
from pathlib import Path

import astropy.units as u
import numpy as np

from simtools.utils.geometry import rotate


class Camera:
    """
    Camera class, defining pixel layout.

    This includes rotation, finding neighbor pixels, calculating FoV and plotting the camera.

    Parameters
    ----------
    telescope_model_name: str
        As provided by the telescope model method TelescopeModel (e.g., LSTN-01)
    camera_config_file: str or Path
        The sim_telarray file name.
    focal_length: float
        The focal length of the camera in (preferably the effective focal length), assumed to be \
        in the same unit as the pixel positions in the camera_config_file, usually cm.
    """

    # Constants for finding neighbor pixels.
    PMT_NEIGHBOR_RADIUS_FACTOR = 1.1
    SIPM_NEIGHBOR_RADIUS_FACTOR = 1.4
    SIPM_ROW_COLUMN_DIST_FACTOR = 0.2

    def __init__(
        self, telescope_model_name: str, camera_config_file: str | Path, focal_length: float
    ):
        """
        Initialize Camera class, defining pixel layout.

        This includes rotation, finding neighbor pixels, calculating FoV and plotting the camera.
        """
        self._logger = logging.getLogger(__name__)

        self.telescope_model_name = telescope_model_name
        self._camera_config_file = camera_config_file
        self.focal_length = focal_length
        if self.focal_length <= 0:
            raise ValueError("The focal length must be larger than zero")
        self.pixels = self.read_pixel_list(camera_config_file)

        self.pixels = self._rotate_pixels(self.pixels)

        # Initialize an empty list of neighbors, to be calculated only when necessary.
        self._neighbors = None

        # Initialize an empty list of edge pixels, to be calculated only when necessary.
        self._edge_pixel_indices = None

    @staticmethod
    def read_pixel_list(camera_config_file: str | Path) -> dict:
        """
        Read the pixel layout from the camera config file, assumed to be in a sim_telarray format.

        Parameters
        ----------
        camera_config_file: str or Path
            The sim_telarray file name.

        Returns
        -------
        dict: pixels
            A dictionary with the pixel positions, the camera rotation angle, the pixel shape, \
            the pixel diameter, the pixel IDs and their "on" status.

        Notes
        -----
        The pixel shape can be hexagonal (denoted as 1 or 3) or a square (denoted as 2). \
        The hexagonal shapes differ in their orientation, where those denoted as 3 are rotated
        clockwise by 30 degrees with respect to those denoted as 1.
        """
        pixels = Camera.initialize_pixel_dict()

        with open(camera_config_file, encoding="utf-8") as dat_file:
            for line in dat_file:
                Camera.process_line(line, pixels)

        Camera.validate_pixels(pixels, camera_config_file)

        return pixels

    @staticmethod
    def initialize_pixel_dict() -> dict:
        """
        Initialize the pixel dictionary with default values.

        Returns
        -------
        dict
            A dictionary with default pixel properties.
        """
        return {
            "pixel_diameter": 9999,
            "pixel_shape": 9999,
            "pixel_spacing": 9999,
            "lightguide_efficiency_angle_file": "none",
            "lightguide_efficiency_wavelength_file": "none",
            "rotate_angle": 0,
            "x": [],
            "y": [],
            "pix_id": [],
            "pix_on": [],
        }

    @staticmethod
    def process_line(line: str, pixels: dict):
        """
        Process a line from the camera config file and update the pixels dictionary.

        Parameters
        ----------
        line: str
            A line from the camera config file.
        pixels: dict
            The dictionary to update with pixel information.
        """
        pix_info = line.split()

        if line.startswith("PixType"):
            pixels["pixel_shape"] = int(pix_info[5].strip())
            pixels["pixel_diameter"] = float(pix_info[6].strip())
            pixels["lightguide_efficiency_angle_file"] = pix_info[8].strip().replace('"', "")

            if len(pix_info) > 9:
                pixels["lightguide_efficiency_wavelength_file"] = (
                    pix_info[9].strip().replace('"', "")
                )

        elif line.startswith("Rotate"):
            pixels["rotate_angle"] = np.deg2rad(float(pix_info[1].strip()))

        elif line.startswith("Pixel"):
            pixels["x"].append(float(pix_info[3].strip()))
            pixels["y"].append(float(pix_info[4].strip()))
            pixels["pix_id"].append(int(pix_info[1].strip()))

            if len(pix_info) > 9:
                pixels["pix_on"].append(int(pix_info[9].strip()) != 0)
            else:
                pixels["pix_on"].append(True)

    @staticmethod
    def validate_pixels(pixels: dict, camera_config_file: str | Path):
        """
        Validate the pixel dictionary to ensure all required fields are present.

        Parameters
        ----------
        pixels: dict
            The pixel dictionary to validate.
        camera_config_file: string
            The sim_telarray file name for error messages.

        Raises
        ------
        ValueError
            If the pixel diameter or pixel shape is invalid.
        """
        if pixels["pixel_diameter"] == 9999:
            raise ValueError(f"Could not read the pixel diameter from {camera_config_file} file")

        if pixels["pixel_shape"] not in [1, 2, 3]:
            raise ValueError(
                f"Pixel shape in {camera_config_file} unrecognized (has to be 1, 2 or 3)"
            )

    def _rotate_pixels(self, pixels: dict) -> dict:
        """
        Rotate the pixels according to the rotation angle given in pixels['rotate_angle'].

        Additional rotation is added to get to the camera view of an observer facing the camera.
        The angle for the axes rotation depends on the coordinate system in which the original
        data was provided.

        Parameters
        ----------
        pixels: dict
            The dictionary produced by the read_pixel_list method of this class

        Returns
        -------
        pixels: dict
            The pixels dictionary with rotated pixels.
            The pixels orientation for plotting is added to the dictionary in pixels['orientation'].
            The orientation is determined by the pixel shape (see read_pixel_list for details).
        """
        rotate_angle = pixels["rotate_angle"] * u.rad  # So not to change the original angle
        # The original pixel list is given such that
        # x -> North, y -> West, z -> Up in the ground system.
        # At azimuth=0, zenith angle=0 all coordinate systems are aligned.
        # When the telescope turns the "normal" way towards
        # the horizon, the x-axis points downwards, the y-axis points right
        # (when looking from the camera onto the dish),
        # and the z-axis points in any case from (primary) dish towards camera.
        # To get the camera for an observer facing the camera, need to rotate by 90 degrees.
        rotate_angle += (90 * u.deg).to(u.rad)

        self._logger.debug(f"Rotating pixels by {rotate_angle.to(u.deg)} (clockwise rotation)")

        if rotate_angle != 0:
            pixels["x"], pixels["y"] = rotate(pixels["x"], pixels["y"], rotate_angle)

        pixels["orientation"] = 0
        if pixels["pixel_shape"] == 1 or pixels["pixel_shape"] == 3:
            if pixels["pixel_shape"] == 3:
                pixels["orientation"] = 30
            if rotate_angle > 0:
                pixels["orientation"] -= rotate_angle.to(u.deg).value

        return pixels

    def get_number_of_pixels(self) -> int:
        """
        Get the number of pixels in the camera (all pixels, including those defined as "off").

        Returns
        -------
        int
            number of pixels.
        """
        return len(self.pixels["x"])

    def get_pixel_diameter(self) -> float:
        """
        Get pixel diameter contained in _pixels.

        Returns
        -------
        float
            Pixel diameter (usually in cm).
        """
        return self.pixels["pixel_diameter"]

    def get_pixel_active_solid_angle(self) -> float:
        """
        Get the active solid angle of a pixel in sr.

        Returns
        -------
        float
            active solid angle of a pixel in sr.
        """
        pixel_area = self.get_pixel_diameter() ** 2
        # In case we have hexagonal pixels:
        if self.get_pixel_shape() == 1 or self.get_pixel_shape() == 3:
            pixel_area *= np.sqrt(3) / 2
        return pixel_area / (self.focal_length**2)

    def get_pixel_shape(self) -> int:
        """
        Get pixel shape code 1, 2 or 3.

        Where 1 and 3 are hexagonal pixels, where one is rotated by\
        30 degrees with respect to the other. A square pixel is denoted as 2.

        Returns
        -------
        int (1, 2 or 3)
            Pixel shape.
        """
        return self.pixels["pixel_shape"]

    def get_lightguide_efficiency_angle_file_name(self) -> str:
        """
        Get the file name of the light guide efficiency as a function of incidence angle.

        Returns
        -------
        str
            File name of the light guide efficiency as a function of incidence angle.
        """
        return self.pixels["lightguide_efficiency_angle_file"]

    def get_lightguide_efficiency_wavelength_file_name(self) -> str:
        """
        Get the file name of the light guide efficiency as a function of wavelength.

        Returns
        -------
        str
            File name of the light guide efficiency as a function of wavelength.
        """
        return self.pixels["lightguide_efficiency_wavelength_file"]

    def get_camera_fill_factor(self) -> float:
        """
        Calculate the fill factor of the camera, defined as (pixel_diameter/pixel_spacing)**2.

        Returns
        -------
        float
            The camera fill factor.
        """
        from scipy.spatial import distance  # pylint: disable=import-outside-toplevel

        if self.pixels["pixel_spacing"] == 9999:
            points = np.array([self.pixels["x"], self.pixels["y"]]).T
            pixel_distances = distance.cdist(points, points, "euclidean")
            # pylint: disable=unsubscriptable-object
            pixel_distances = pixel_distances[pixel_distances > 0]
            pixel_spacing = np.min(pixel_distances)
            self.pixels["pixel_spacing"] = pixel_spacing

        return (self.pixels["pixel_diameter"] / self.pixels["pixel_spacing"]) ** 2

    def calc_fov(self) -> tuple[float, float]:
        """
        Calculate the FOV of the camera in degrees, taking into account the focal length.

        Returns
        -------
        fov: float
            The FOV of the camera in the degrees.
        average_edge_distance: float
            The average edge distance of the camera.

        Notes
        -----
        The x,y pixel positions and focal length are assumed to have the same unit (usually cm)
        """
        self._logger.debug("Calculating the FoV")
        return self._calc_fov(
            self.pixels["x"],
            self.pixels["y"],
            self.get_edge_pixels(),
            self.focal_length,
        )

    def _calc_fov(
        self,
        x_pixel: list[float],
        y_pixel: list[float],
        edge_pixel_indices: list[int],
        focal_length: float,
    ) -> tuple[float, float]:
        """
        Calculate the FOV of the camera in degrees, taking into account the focal length.

        Parameters
        ----------
        x_pixel: list
            List of positions of the pixels on the x-axis
        y_pixel: list
            List of positions of the pixels on the y-axis
        edge_pixel_indices: list
            List of indices of the edge pixels
        focal_length: float
            The focal length of the camera in (preferably the effective focal length), assumed to \
            be in the same unit as the pixel positions.

        Returns
        -------
        fov: float
            The FOV of the camera in the degrees.
        average_edge_distance: float
            The average edge distance of the camera

        Notes
        -----
        The x,y pixel positions and focal length are assumed to have the same unit (usually cm)
        """
        self._logger.debug("Calculating the FoV")

        average_edge_distance = 0
        for i_pix in edge_pixel_indices:
            average_edge_distance += np.sqrt(x_pixel[i_pix] ** 2 + y_pixel[i_pix] ** 2)
        average_edge_distance /= len(edge_pixel_indices)

        fov = 2 * np.rad2deg(np.arctan(average_edge_distance / focal_length))

        return fov, average_edge_distance

    @staticmethod
    def _find_neighbors(x_pos: np.ndarray, y_pos: np.ndarray, radius: float) -> list[list[int]]:
        """
        Use a KD-Tree to quickly find nearest neighbors.

        This applies to e.g., of the pixels in a camera or mirror facets.

        Parameters
        ----------
        x_pos : numpy.array_like
            x position of each e.g., pixel
        y_pos : numpy.array_like
            y position of each e.g., pixel
        radius : float
            radius to consider neighbor it should be slightly larger than the pixel diameter or \
            mirror facet.

        Returns
        -------
        list of lists
            Array of neighbor indices in a list for each pixel
        """
        from scipy.spatial import cKDTree as KDTree  # pylint: disable=import-outside-toplevel

        tree = KDTree(np.column_stack([x_pos, y_pos]))
        neighbors = tree.query_ball_tree(tree, radius)
        return [list(np.setdiff1d(neigh, [i])) for i, neigh in enumerate(neighbors)]

    def _find_adjacent_neighbor_pixels(
        self, x_pos: np.ndarray, y_pos: np.ndarray, radius: float, row_column_dist: float
    ) -> list[list[int]]:
        """
        Find adjacent neighbor pixels in cameras with square pixels.

        Only directly adjacent neighbors are allowed, no diagonals.

        Parameters
        ----------
        x_pos: np.ndarray
            x position of each pixel
        y_pos: np.ndarray
            y position of each pixel
        radius: float
            Radius within which to find neighbors
        row_column_dist: float
            Distance to consider for row/column adjacency.
            Should be around 20% of the pixel diameter.

        Returns
        -------
        list of lists
            Array of neighbor indices in a list for each pixel
        """
        # First find the neighbors with the usual method and the original radius
        # which does not allow for diagonal neighbors.
        neighbors = self._find_neighbors(x_pos, y_pos, radius)

        for i_pix, nn in enumerate(neighbors):
            # Find pixels defined as edge pixels now
            if len(nn) < 4:
                # Go over all other pixels and search for ones which are adjacent
                # but further than sqrt(2) away
                self._add_additional_neighbors(i_pix, nn, x_pos, y_pos, radius, row_column_dist)

        return neighbors

    def _add_additional_neighbors(
        self,
        i_pix: int,
        nn: list[int],
        x_pos: np.ndarray,
        y_pos: np.ndarray,
        radius: float,
        row_column_dist: float,
    ):
        """
        Add neighbors for a given pixel if they are not already neighbors and are adjacent.

        Parameters
        ----------
        i_pix: int
            Index of the pixel to find neighbors for
        nn: list
            Current list of neighbors for the pixel
        x_pos: np.ndarray
            x position of each pixel
        y_pos: np.ndarray
            y position of each pixel
        radius: float
            Radius within which to find neighbors
        row_column_dist: float
            Distance to consider for row/column adjacency
        """
        for j_pix, _ in enumerate(x_pos):
            # No need to look at the pixel itself
            # nor at any pixels already in the neighbors list
            if j_pix != i_pix and j_pix not in nn:
                dist = np.sqrt(
                    (x_pos[i_pix] - x_pos[j_pix]) ** 2 + (y_pos[i_pix] - y_pos[j_pix]) ** 2
                )
                # Check if this pixel is in the same row or column
                # and allow it to be ~1.68*diameter away (1.4*1.2 = 1.68)
                # Need to increase the distance because of the curvature
                # of the CHEC camera
                if (
                    abs(x_pos[i_pix] - x_pos[j_pix]) < row_column_dist
                    or abs(y_pos[i_pix] - y_pos[j_pix]) < row_column_dist
                ) and dist < 1.2 * radius:
                    nn.append(j_pix)

    def _calc_neighbor_pixels(self, pixels: dict) -> list[list[int]]:
        """
        Find adjacent neighbor pixels in cameras with hexagonal or square pixels.

        Only directly adjacent neighbors are searched for, no diagonals.

        Parameters
        ----------
        pixels: dict
            The dictionary produced by the read_pixel_list method of this class

        Returns
        -------
        neighbors: list of lists
            Array of neighbor indices in a list for each pixel
        """
        self._logger.debug("Searching for neighbor pixels")

        if pixels["pixel_shape"] == 1 or pixels["pixel_shape"] == 3:
            self._neighbors = self._find_neighbors(
                pixels["x"],
                pixels["y"],
                self.PMT_NEIGHBOR_RADIUS_FACTOR * pixels["pixel_diameter"],
            )
        elif pixels["pixel_shape"] == 2:
            # Distance increased by 40% to take into account gaps in the SiPM cameras
            # Pixels in the same row/column can be 20% shifted from one another
            # Inside find_adjacent_neighbor_pixels the distance is increased
            # further for pixels in the same row/column to 1.68*diameter.
            self._neighbors = self._find_adjacent_neighbor_pixels(
                pixels["x"],
                pixels["y"],
                self.SIPM_NEIGHBOR_RADIUS_FACTOR * pixels["pixel_diameter"],
                self.SIPM_ROW_COLUMN_DIST_FACTOR * pixels["pixel_diameter"],
            )

        return self._neighbors

    def get_neighbor_pixels(self, pixels: dict | None = None) -> list[list[int]]:
        """
        Get a list of neighbor pixels by calling calc_neighbor_pixels() when necessary.

        The purpose of this function is to ensure the calculation occurs only once and only when
        necessary.

        Parameters
        ----------
        pixels: dict
            The dictionary produced by the read_pixel_list method of this class.

        Returns
        -------
        neighbors: list of lists
            Array of neighbor indices in a list for each pixel.
        """
        if self._neighbors is None:
            if pixels is None:
                pixels = self.pixels
            return self._calc_neighbor_pixels(pixels)

        return self._neighbors

    def _calc_edge_pixels(self, pixels: dict, neighbors: list[list[int]]) -> list[int]:
        """
        Find the edge pixels of the camera.

        Parameters
        ----------
        pixels: dict
            The dictionary produced by the read_pixel_list method of this class.
        neighbors: list of lists
            Array of neighbor indices in a list for each pixel.

        Returns
        -------
        edge_pixel_indices: list
            Array of edge pixel indices.
        """
        self._logger.debug("Searching for edge pixels")
        edge_pixel_indices = []

        def is_edge_pixel(i_pix):
            pixel_shape = pixels["pixel_shape"]
            pix_on = pixels["pix_on"][i_pix]
            num_neighbors = len(neighbors[i_pix])

            shape_condition = (pixel_shape in [1, 3] and num_neighbors < 6) or (
                pixel_shape == 2 and num_neighbors < 4
            )
            return pix_on and shape_condition

        for i_pix, _ in enumerate(pixels["x"]):
            if is_edge_pixel(i_pix):
                edge_pixel_indices.append(i_pix)

        return edge_pixel_indices

    def get_edge_pixels(
        self, pixels: dict | None = None, neighbors: list[list[int]] | None = None
    ) -> list[int]:
        """
        Get the indices of the edge pixels of the camera.

        Parameters
        ----------
        pixels: dict
            The dictionary produced by the read_pixel_list method of this class.
        neighbors: list of lists
            Array of neighbor indices in a list for each pixel.

        Returns
        -------
        edge_pixel_indices: list
            Array of edge pixel indices.
        """
        if self._edge_pixel_indices is None:
            if pixels is None:
                pixels = self.pixels
            if neighbors is None:
                neighbors = self.get_neighbor_pixels()
            return self._calc_edge_pixels(pixels, neighbors)

        return self._edge_pixel_indices



#!/usr/bin/python3
"""Reading of legacy data files (expect that this will be obsolete in future)."""

import logging

from astropy.table import Table

logger = logging.getLogger(__name__)


def read_legacy_data_as_table(file_path, file_type):
    """
    Read legacy data file.

    Parameters
    ----------
    file_path: Path
        Path to the legacy data file.
    file_type: str
        Type of legacy data file.

    Returns
    -------
    Table
        Astropy table.

    Raises
    ------
    ValueError
        If unsupported legacy data file type.
    """
    logger.debug(f"Reading legacy data file of type {file_type} from {file_path}")

    try:
        return globals()[f"read_{file_type}"](file_path)
    except KeyError as exc:
        raise ValueError(f"Unsupported legacy data file type: {file_type}") from exc


def read_legacy_lst_single_pe(file_path):
    """
    Read LST single pe file (in legacy data format).

    File contains two columns: amplitude (in units of single p.e) and response.

    Parameters
    ----------
    file_path: Path
        Path to the legacy data file.

    Returns
    -------
    Table
        Astropy table.
    """
    return Table.read(file_path, format="ascii.csv", names=("amplitude", "response"))



#!/usr/bin/python3

r"""
    Add a file to a DB.

    The name and location of the file are required.
    This application should complement the ones for updating parameters, \
    adding entries to the DB and getting files from the DB.

    Command line arguments
    ----------------------
    file_name (str or list of str, required)
        Name of the file to upload including the full path. \
        A list of files is also allowed, in which case only one -file_name is necessary, \
        i.e., python applications/db_add_file_to_db.py -file_name file_1.dat file_2.dat file_3.dat \
        If no path is given, the file is assumed to be in the CWD.
    input_path (str, required if file_name is not given)
        A directory with files to upload to the DB. \
        All files in the directory with a predefined list of extensions will be uploaded.
    db (str)
        The DB to insert the files to.

    Example
    -------
    uploading a dummy file.

    .. code-block:: console

        simtools-db-add-file-to-db --file_name test_application.dat --db test-data

    Expected final print-out message:

    .. code-block:: console

        INFO::get_file_from_db(l75)::main::Got file test_application.dat from DB test-data and
        saved into .

"""

import uuid
from pathlib import Path

import simtools.utils.general as gen
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.db import db_handler


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Add file to the DB.",
        usage="simtools-add-file-to-db --file_name test_application.dat --db test-data",
    )
    group = config.parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--file_name",
        help=("The file name to upload. A list of files is also allowed."),
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--input_path",
        help=("A directory with files to upload to the DB."),
        type=Path,
    )

    config.parser.add_argument(
        "--db",
        type=str,
        help=("The database to insert the files to."),
    )
    config.parser.add_argument(
        "--test_db",
        help="Use sandbox database. Drop all data after the operation.",
        action="store_true",
    )

    return config.initialize(paths=False, db_config=True)


def collect_files_to_insert(args_dict, logger, db):
    """
    Collect the files to insert into the database based on the provided arguments.

    Parameters
    ----------
    args_dict : dict
        Dictionary of parsed command-line arguments.
    logger : logging.Logger
        Logger object for logging messages.
    db : DatabaseHandler
        Database handler object.

    Returns
    -------
    list
        List of files to be inserted into the database.

    Raises
    ------
    ValueError
        If no valid files are provided for uploading.
    """
    files_to_insert = []

    if args_dict.get("file_name", None) is not None:
        for file_now in args_dict["file_name"]:
            if Path(file_now).suffix in db.ALLOWED_FILE_EXTENSIONS:
                files_to_insert.append(file_now)
            else:
                logger.warning(
                    f"The file {file_now} will not be uploaded to the DB because its extension "
                    f"is not in the allowed extension list: {db.ALLOWED_FILE_EXTENSIONS}"
                )
    else:
        for ext_now in db.ALLOWED_FILE_EXTENSIONS:
            files_to_insert.extend(Path(args_dict["input_path"]).glob(f"*{ext_now}"))

    if not files_to_insert:
        raise ValueError("No files were provided to upload")

    return files_to_insert


def confirm_and_insert_files(files_to_insert, args_dict, db, logger):
    """
    Confirm the files to be inserted and insert them into the database.

    Parameters
    ----------
    files_to_insert : list
        List of files to be inserted into the database.
    args_dict : dict
        Dictionary of parsed command-line arguments.
    db : DatabaseHandler
        Database handler object.
    logger : logging.logger
        logger object for logging messages.
    """
    plural = "" if len(files_to_insert) == 1 else "s"

    if args_dict.get("test_db", False):
        args_dict["db"] = args_dict["db"] + str(uuid.uuid4())
        logger.info(f"Using test database: {args_dict['db']}")

    print(f"Should the following file{plural} be inserted to the {args_dict['db']} DB?:\n")
    print(*files_to_insert, sep="\n")
    print()

    if gen.user_confirm():
        for file_to_insert_now in files_to_insert:
            db.insert_file_to_db(file_to_insert_now, args_dict["db"])
            logger.info(f"File {file_to_insert_now} inserted to {args_dict['db']} DB")
    else:
        logger.info(f"Aborted, did not insert file{plural} to the {args_dict['db']} DB")

    # drop test database; be safe and required DB name is sandbox
    if args_dict.get("test_db", False) and "sandbox" in args_dict["db"]:
        logger.info(f"Test database used. Dropping all data from {args_dict['db']}")
        db.mongo_db_handler.db_client.drop_database(args_dict["db"])


def main():
    """Add files to the database."""
    app_context = startup_application(_parse, setup_io_handler=False)

    db = db_handler.DatabaseHandler(db_config=app_context.db_config)

    files_to_insert = collect_files_to_insert(app_context.args, app_context.logger, db)
    confirm_and_insert_files(files_to_insert, app_context.args, db, app_context.logger)


if __name__ == "__main__":
    main()



"""Functions asserting certain conditions are met (used e.g., in integration tests)."""

import json
import logging
from collections import defaultdict
from pathlib import Path

import numpy as np
import yaml

from simtools.simtel.simtel_io_metadata import read_sim_telarray_metadata

_logger = logging.getLogger(__name__)


def assert_file_type(file_type, file_name):
    """
    Assert that the file is of the given type.

    Parameters
    ----------
    file_type: str
        File type (json, yaml).
    file_name: str
        File name.

    """
    if file_type == "json":
        try:
            with open(file_name, encoding="utf-8") as file:
                json.load(file)
            return True
        except (json.JSONDecodeError, FileNotFoundError):
            return False
    if file_type in ("yaml", "yml"):
        if Path(file_name).suffix[1:] not in ("yaml", "yml"):
            return False
        try:
            with open(file_name, encoding="utf-8") as file:
                yaml.safe_load(file)
            return True
        except (yaml.YAMLError, FileNotFoundError):
            return False

    # no dedicated tests for other file types, checking suffix only
    _logger.info(f"File type test is checking suffix only for {file_name} (suffix: {file_type}))")
    return Path(file_name).suffix[1:] == file_type


def assert_n_showers_and_energy_range(file):
    """
    Assert the number of showers and the energy range.

    The number of showers should be consistent with the required one (up to 1% tolerance)
    and the energies simulated are required to be within the configured ones.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.

    """
    from eventio.simtel.simtelfile import SimTelFile  # pylint: disable=import-outside-toplevel

    simulated_energies = []
    simulation_config = {}
    with SimTelFile(file, skip_non_triggered=False) as f:
        simulation_config = f.mc_run_headers[0]
        for event in f:
            simulated_energies.append(event["mc_shower"]["energy"])

    # The relative tolerance is set to 1% because ~0.5% shower simulations do not
    # succeed, without resulting in an error. This tolerance therefore is not an issue.
    consistent_n_showers = np.isclose(
        len(np.unique(simulated_energies)), simulation_config["n_showers"], rtol=1e-2
    )
    consistent_energy_range = all(
        simulation_config["E_range"][0] <= energy <= simulation_config["E_range"][1]
        for energy in simulated_energies
    )

    return consistent_n_showers and consistent_energy_range


def assert_expected_output(file, expected_output):
    """
    Assert that the expected output is present in the sim_telarray file.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.
    expected_output: dict
        Expected output values.

    """
    from eventio.simtel.simtelfile import SimTelFile  # pylint: disable=import-outside-toplevel

    item_to_check = defaultdict(list)
    with SimTelFile(file) as f:
        for event in f:
            if "pe_sum" in expected_output:
                item_to_check["pe_sum"].extend(
                    event["photoelectron_sums"]["n_pe"][event["photoelectron_sums"]["n_pe"] > 0]
                )
            if "trigger_time" in expected_output:
                item_to_check["trigger_time"].extend(event["trigger_information"]["trigger_times"])
            if "photons" in expected_output:
                item_to_check["photons"].extend(
                    event["photoelectron_sums"]["photons_atm_qe"][
                        event["photoelectron_sums"]["photons"] > 0
                    ]
                )

    for key, value in expected_output.items():
        if len(item_to_check[key]) == 0:
            _logger.error(f"No data found for {key}")
            return False

        if not value[0] < np.mean(item_to_check[key]) < value[1]:
            _logger.error(
                f"Mean of {key} is not in the expected range, got {np.mean(item_to_check[key])}"
            )
            return False

    return True


def assert_expected_simtel_metadata(file, expected_metadata):
    """
    Assert that expected metadata is present in the sim_telarray file.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.
    expected_metadata: dict
        Expected metadata values.

    """
    global_meta, telescope_meta = read_sim_telarray_metadata(file)

    for key, value in expected_metadata.items():
        if key not in global_meta and key not in telescope_meta:
            _logger.error(f"Metadata key {key} not found in sim_telarray file {file}")
            return False
        if key in global_meta and global_meta[key] != value:
            _logger.error(
                f"Metadata key {key} has value {global_meta[key]} instead of expected {value}"
            )
            return False
        _logger.debug(f"Metadata key {key} matches expected value {value}")

    return True


def check_output_from_sim_telarray(file, file_test):
    """
    Check that the sim_telarray simulation result is reasonable and matches the expected output.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.
    file_test: dict
        File test description including expected output and metadata.

    Raises
    ------
    ValueError
        If the file is not a zstd compressed file.
    """
    if "expected_output" not in file_test and "expected_simtel_metadata" not in file_test:
        _logger.debug(f"No expected output or metadata provided, skipping checks {file_test}")
        return True

    if file.suffix != ".zst":
        raise ValueError(
            f"Expected output file {file} is not a zstd compressed file "
            f"(i.e., a sim_telarray file)."
        )

    assert_output = assert_metadata = True

    if "expected_output" in file_test:
        assert_output = assert_expected_output(
            file=file, expected_output=file_test["expected_output"]
        )

    if "expected_simtel_metadata" in file_test:
        assert_metadata = assert_expected_simtel_metadata(
            file=file, expected_metadata=file_test["expected_simtel_metadata"]
        )

    return assert_n_showers_and_energy_range(file=file) and assert_output and assert_metadata



#!/usr/bin/python3
"""Read file info and run headers from sim_telarray files."""

from eventio import EventIOFile
from eventio.simtel import MCRunHeader, MCShower, RunHeader


def get_corsika_run_number(file):
    """
    Return the CORSIKA run number from a sim_telarray file.

    Parameters
    ----------
    file: str
        Path to the sim_telarray file.

    Returns
    -------
    int, None
        CORSIKA run number. Returns None if not found.
    """
    run_header = get_corsika_run_header(file)
    return run_header.get("run") if run_header else None


def get_corsika_run_header(file):
    """
    Return the CORSIKA run header information from a sim_telarray file.

    Reads both RunHeader and MCRunHeader object from file and
    returns a merged dictionary. Adds primary id from the first event.

    Parameters
    ----------
    file: str
        Path to the sim_telarray file.

    Returns
    -------
    dict, None
        CORSIKA run header. Returns None if not found.
    """
    run_header = None
    mc_run_header = None
    primary_id = None

    with EventIOFile(file) as f:
        for o in f:
            if isinstance(o, RunHeader) and run_header is None:
                run_header = o.parse()
            elif isinstance(o, MCRunHeader) and mc_run_header is None:
                mc_run_header = o.parse()
            elif isinstance(o, MCShower):  # get primary_id from first MCShower
                primary_id = o.parse().get("primary_id")
            if run_header and mc_run_header and primary_id is not None:
                break

    run_header = run_header or {}
    mc_run_header = mc_run_header or {}
    if primary_id is not None:
        mc_run_header["primary_id"] = primary_id
    return run_header | mc_run_header or None



"""Read reduced datasets in form of astropy tables from file."""

import logging
from dataclasses import dataclass, field

import astropy.units as u
import numpy as np
from astropy.coordinates import angular_separation

from simtools.corsika.primary_particle import PrimaryParticle
from simtools.io import table_handler
from simtools.utils.geometry import solid_angle, transform_ground_to_shower_coordinates


@dataclass
class ShowerEventData:
    """Container for shower event data."""

    shower_id: list[np.uint32] = field(default_factory=list)
    event_id: list[np.uint32] = field(default_factory=list)
    file_id: list[np.uint32] = field(default_factory=list)
    simulated_energy: list[np.float64] = field(default_factory=list)
    x_core: list[np.float64] = field(default_factory=list)
    y_core: list[np.float64] = field(default_factory=list)
    shower_azimuth: list[np.float64] = field(default_factory=list)
    shower_altitude: list[np.float64] = field(default_factory=list)
    area_weight: list[np.float64] = field(default_factory=list)
    x_core_shower: list[np.float64] = field(default_factory=list)
    y_core_shower: list[np.float64] = field(default_factory=list)
    core_distance_shower: list[np.float64] = field(default_factory=list)
    angular_distance: list[float] = field(default_factory=list)


@dataclass
class TriggeredEventData:
    """Container for triggered event data."""

    shower_id: list[np.uint32] = field(default_factory=list)
    event_id: list[np.uint32] = field(default_factory=list)
    file_id: list[np.uint32] = field(default_factory=list)
    array_altitude: list[float] = field(default_factory=list)
    array_azimuth: list[float] = field(default_factory=list)
    telescope_list: list[np.ndarray] = field(default_factory=list)
    angular_distance: list[float] = field(default_factory=list)


class SimtelIOEventDataReader:
    """Read reduced MC data set stored in astropy tables."""

    def __init__(self, event_data_file, telescope_list=None):
        """Initialize SimtelIOEventDataReader."""
        self._logger = logging.getLogger(__name__)
        self.telescope_list = telescope_list

        self.data_sets = self.read_table_list(event_data_file)
        self.reduced_file_info = None

    def read_table_list(self, event_data_file):
        """
        Read available tables from the event data file.

        Rearrange dictionary with tables names into a list of dictionaries
        under the assumption that the file contains the tables "SHOWERS",
        "TRIGGERS", and "FILE_INFO".

        Parameters
        ----------
        event_data_file : str
            Path to the event data file.

        Returns
        -------
        list
            List of dictionaries containing the data from the tables.
        """
        dataset_dict = table_handler.read_table_list(
            event_data_file,
            ["SHOWERS", "TRIGGERS", "FILE_INFO"],
            include_indexed_tables=True,
        )

        data_sets = []
        try:
            sorted_indices = sorted(
                range(len(dataset_dict["SHOWERS"])),
                key=lambda i: int(dataset_dict["SHOWERS"][i].split("_")[-1]),
            )
        except (ValueError, AttributeError):
            sorted_indices = [0]  # Handle the case where the key is only "SHOWERS"
        for i in sorted_indices:
            data_sets.append(
                {
                    "SHOWERS": dataset_dict["SHOWERS"][i],
                    "TRIGGERS": dataset_dict["TRIGGERS"][i],
                    "FILE_INFO": dataset_dict["FILE_INFO"][i],
                }
            )

        return data_sets

    def _table_to_shower_data(self, table):
        """
        Convert tables with shower event data and add derived quantities.

        Parameters
        ----------
        table : astropy.table.Table
            Table "SHOWERS"

        Returns
        -------
        ShowerEventData
            An instance of ShowerEventData with the data from the table.
        """
        shower_data = ShowerEventData()

        for col in table.colnames:
            setattr(shower_data, col, np.array(table[col].data))
            if table[col].unit:
                setattr(shower_data, f"{col}_unit", table[col].unit)

        shower_data.x_core_shower, shower_data.y_core_shower, _ = (
            transform_ground_to_shower_coordinates(
                shower_data.x_core,
                shower_data.y_core,
                0.0,
                shower_data.shower_azimuth,
                shower_data.shower_altitude,
            )
        )
        shower_data.core_distance_shower = np.hypot(
            shower_data.x_core_shower, shower_data.y_core_shower
        )
        shower_data.angular_distance = (
            angular_separation(
                shower_data.shower_azimuth * u.deg,
                shower_data.shower_altitude * u.deg,
                self.reduced_file_info["azimuth"],
                (90.0 * u.deg - self.reduced_file_info["zenith"]),
            )
            .to(u.deg)
            .value
        )

        return shower_data

    def _table_to_triggered_data(self, table):
        """
        Convert table with triggered event data.

        Converts telescope lists from comma-separated string to numpy array.

        Parameters
        ----------
        table : astropy.table.Table
            Table "TRIGGERS"

        Returns
        -------
        TriggeredEventData
            An instance of TriggeredEventData with the data from the table.
        """
        triggered_data = TriggeredEventData()
        for col in table.colnames:
            if col == "telescope_list":
                arrays = [
                    np.array(list(map(str, tel_list.split(","))), dtype=np.str_)
                    for tel_list in table[col]
                ]
                triggered_data.telescope_list = arrays
            else:
                data = np.array(table[col].data)
                setattr(triggered_data, col, data)
                if table[col].unit:
                    setattr(triggered_data, f"{col}_unit", table[col].unit)
        return triggered_data

    def _get_triggered_shower_data(
        self, shower_data, triggered_file_id, triggered_event_id, triggered_shower_id
    ):
        """
        Get shower data corresponding to triggered events.

        Matches triggered events with showers based on shower_id, event_id, and file_id.

        Parameters
        ----------
        shower_data : ShowerEventData
            The shower data containing all showers.
        triggered_file_id : list
            List of file IDs for triggered events.
        triggered_event_id : list
            List of event IDs for triggered events.
        triggered_shower_id : list
            List of shower IDs for triggered events.

        Returns
        -------
        ShowerEventData
            An instance of ShowerEventData containing only the triggered showers.

        """
        triggered_shower = ShowerEventData()

        matched_indices = []
        for tr_shower_id, tr_event_id, tr_file_id in zip(
            triggered_shower_id, triggered_event_id, triggered_file_id
        ):
            mask = (
                (shower_data.shower_id == tr_shower_id)
                & (shower_data.event_id == tr_event_id)
                & (shower_data.file_id == tr_file_id)
            )
            matched_idx = np.nonzero(mask)[0]
            if len(matched_idx) == 1:
                matched_indices.append(matched_idx[0])
            else:
                self._logger.warning(
                    f"Found {len(matched_idx)} matches for shower {tr_shower_id}"
                    f" event {tr_event_id} file {tr_file_id}"
                )

        for attr in vars(shower_data):
            if not attr.endswith("_unit"):
                value = getattr(shower_data, attr)
                if isinstance(value, list | np.ndarray):
                    setattr(triggered_shower, attr, np.array(value)[matched_indices])

        return triggered_shower

    def read_event_data(self, event_data_file, table_name_map=None):
        """
        Read event data and file info tables from file and apply transformations.

        Allows to map tables names to their actual names in the file
        (e.g., "SHOWER" to "SHOWER_1").

        Parameters
        ----------
        event_data_file : str
            Path to the event data file.
        table_name_map : dict, optional
            Mapping of table names to their actual names in the file.
            Defaults to using the standard names "SHOWERS", "TRIGGERS", and "FILE_INFO".

        Returns
        -------
        tuple
            A tuple with file info table, shower, triggered shower, and triggered event data.
        """
        table_name_map = table_name_map or {}

        def get_name(key):
            return table_name_map.get(key, key)

        tables = table_handler.read_tables(
            event_data_file,
            table_names=[get_name(k) for k in ("SHOWERS", "TRIGGERS", "FILE_INFO")],
        )
        self.reduced_file_info = self.get_reduced_simulation_file_info(
            tables[get_name("FILE_INFO")]
        )

        shower_data = self._table_to_shower_data(tables[get_name("SHOWERS")])
        triggered_data = self._table_to_triggered_data(tables[get_name("TRIGGERS")])
        triggered_shower = self._get_triggered_shower_data(
            shower_data,
            tables[get_name("TRIGGERS")]["file_id"],
            tables[get_name("TRIGGERS")]["event_id"],
            tables[get_name("TRIGGERS")]["shower_id"],
        )

        triggered_data.angular_distance = (
            angular_separation(
                triggered_shower.shower_azimuth * u.deg,
                triggered_shower.shower_altitude * u.deg,
                triggered_data.array_azimuth * u.deg,
                triggered_data.array_altitude * u.deg,
            )
            .to(u.deg)
            .value
        )

        if self.telescope_list:
            triggered_data, triggered_shower = self._filter_by_telescopes(
                triggered_data, triggered_shower
            )

        self._logger.info(f"Number of triggered events: {len(triggered_data.array_altitude)}")

        return (
            tables[get_name("FILE_INFO")],
            shower_data,
            triggered_shower,
            triggered_data,
        )

    def _filter_by_telescopes(self, triggered_data, triggered_shower):
        """Filter trigger data and triggered shower data by the specified telescope list."""
        mask = np.array(
            [
                any(tel in event for tel in self.telescope_list)
                for event in triggered_data.telescope_list
            ]
        )
        filtered_triggered_data = TriggeredEventData(
            shower_id=triggered_data.shower_id[mask],
            event_id=triggered_data.event_id[mask],
            file_id=triggered_data.file_id[mask],
            array_altitude=triggered_data.array_altitude[mask],
            array_azimuth=triggered_data.array_azimuth[mask],
            telescope_list=[triggered_data.telescope_list[i] for i in np.arange(len(mask))[mask]],
            angular_distance=triggered_data.angular_distance[mask],
        )
        filtered_triggered_shower_data = self._get_triggered_shower_data(
            triggered_shower,
            filtered_triggered_data.file_id,
            filtered_triggered_data.event_id,
            filtered_triggered_data.shower_id,
        )

        return filtered_triggered_data, filtered_triggered_shower_data

    def get_reduced_simulation_file_info(self, simulation_file_info):
        """
        Return reduced simulation file info assuming single-valued parameters.

        Applies rounding and uniqueness functions extract representative values
        for zenith, azimuth, and NSB level. Assumes all files share identical
        simulation parameters except for file names. Returns particle name instead
        of ID.

        Logs a warning if multiple unique values are found.

        Parameters
        ----------
        simulation_file_info : astropy.table.Table
            Dictionary containing simulation file info.

        Returns
        -------
        dict
            Dictionary containing the reduced simulation file info.
        """
        particle_id = np.unique(simulation_file_info["particle_id"].data)
        keys = [
            "zenith",
            "azimuth",
            "nsb_level",
            "energy_min",
            "energy_max",
            "viewcone_min",
            "viewcone_max",
            "core_scatter_min",
            "core_scatter_max",
        ]
        float_arrays = {}
        for key in keys:
            if key == "energy_min":
                float_arrays[key] = np.unique(np.round(simulation_file_info[key].data, decimals=3))
            else:
                float_arrays[key] = np.unique(np.round(simulation_file_info[key].data, decimals=2))

        if any(len(arr) > 1 for arr in (particle_id, *(float_arrays[key] for key in keys))):
            self._logger.warning("Simulation file info has non-unique values.")

        reduced_info = {
            "primary_particle": PrimaryParticle(
                particle_id_type="corsika7_id",
                particle_id=int(particle_id[0]),
            ).name,
        }

        for key in keys:
            value = float(float_arrays[key][0])
            if simulation_file_info[key].unit is not None:
                value = value * simulation_file_info[key].unit
            reduced_info[key] = value

        reduced_info["solid_angle"] = solid_angle(
            angle_min=reduced_info.get("viewcone_min", 0.0 * u.rad),
            angle_max=reduced_info.get("viewcone_max", 0.0 * u.rad),
        )
        reduced_info["scatter_area"] = self.scatter_area(
            core_scatter_min=reduced_info.get("core_scatter_min", 0.0 * u.m),
            core_scatter_max=reduced_info.get("core_scatter_max", 0.0 * u.m),
        )

        return reduced_info

    def scatter_area(self, core_scatter_min, core_scatter_max):
        """
        Calculate the scatter area of the core.

        Parameters
        ----------
        core_scatter_min : astropy.units.Quantity
            Minimum core scatter radius.
        core_scatter_max : astropy.units.Quantity
            Maximum core scatter radius.

        Returns
        -------
        astropy.units.Quantity
            Scatter area.
        """
        return np.pi * (core_scatter_max**2 - core_scatter_min**2)



"""Single photon electron spectral analysis."""

import logging
import re
import subprocess
import tempfile
from io import BytesIO
from pathlib import Path

import numpy as np
from astropy.table import Table
from scipy.optimize import curve_fit

import simtools.data_model.model_data_writer as writer
from simtools.constants import MODEL_PARAMETER_SCHEMA_URL, SCHEMA_PATH
from simtools.data_model import validate_data
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.io import io_handler


class SinglePhotonElectronSpectrum:
    """
    Single photon electron spectral analysis.

    Parameters
    ----------
    args_dict: dict
        Dictionary with input arguments.
    """

    prompt_column = "frequency (prompt)"
    prompt_plus_afterpulse_column = "frequency (prompt+afterpulsing)"
    afterpulse_column = "frequency (afterpulsing)"

    input_schema = SCHEMA_PATH / "input" / "single_pe_spectrum.schema.yml"

    def __init__(self, args_dict):
        """Initialize SinglePhotonElectronSpectrum class."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Initialize SinglePhotonElectronSpectrum class.")

        self.args_dict = args_dict
        # default output is of ecsv format
        self.args_dict["output_file"] = str(
            Path(self.args_dict["output_file"]).with_suffix(".ecsv")
        )
        self.io_handler = io_handler.IOHandler()
        self.data = ""  # Single photon electron spectrum data (as string)
        self.args_dict["metadata_product_data_name"] = "single_pe_spectrum"
        self.args_dict["metadata_product_data_url"] = (
            MODEL_PARAMETER_SCHEMA_URL + "/pm_photoelectron_spectrum.schema.yml"
        )
        self.metadata = MetadataCollector(args_dict=self.args_dict)

    def derive_single_pe_spectrum(self):
        """Derive single photon electron spectrum."""
        afterpulse_fitted_spectrum = (
            self.fit_afterpulse_spectrum() if self.args_dict.get("fit_afterpulse") else None
        )

        if self.args_dict.get("use_norm_spe"):
            return self._derive_spectrum_norm_spe(
                input_spectrum=self.args_dict["input_spectrum"],
                afterpulse_spectrum=self.args_dict.get("afterpulse_spectrum"),
                afterpulse_fitted_spectrum=afterpulse_fitted_spectrum,
            )

        raise NotImplementedError(
            "Derivation of single photon electron spectrum using a simtool is not yet implemented."
        )

    def write_single_pe_spectrum(self):
        """
        Write single photon electron spectrum plus metadata to disk.

        Includes writing in simtel and simtools (ecsv) formats.

        """
        simtel_file = self.io_handler.get_output_directory() / Path(
            self.args_dict["output_file"]
        ).with_suffix(".dat")
        self._logger.debug(f"norm_spe output file: {simtel_file}")
        with open(simtel_file, "w", encoding="utf-8") as simtel:
            simtel.write(self.data)

        cleaned_data = re.sub(r"%%%.+", "", self.data)  # remove norm_spe row metadata
        table = Table.read(
            BytesIO(cleaned_data.encode("utf-8")),
            format="ascii.no_header",
            comment="#",
            delimiter="\t",
        )
        table.rename_columns(
            ["col1", "col2", "col3"],
            ["amplitude", self.prompt_column, self.prompt_plus_afterpulse_column],
        )

        writer.ModelDataWriter.dump(
            args_dict=self.args_dict,
            metadata=self.metadata,
            product_data=table,
            validate_schema_file=None,
        )

    def _derive_spectrum_norm_spe(
        self, input_spectrum, afterpulse_spectrum, afterpulse_fitted_spectrum
    ):
        """
        Derive single photon electron spectrum using sim_telarray tool 'norm_spe'.

        Parameters
        ----------
        input_spectrum : str
            Input file with amplitude spectrum
            (prompt spectrum only if afterpulse spectrum is given).
        afterpulse_spectrum : str
            Input file with afterpulse spectrum.
        afterpulse_fitted_spectrum : astro.Table
            Fitted afterpulse spectrum data.

        Returns
        -------
        int
            Return code of the executed command

        Raises
        ------
        subprocess.CalledProcessError
            If the command execution fails.
        """
        tmp_input_file = self._get_input_data(
            input_file=input_spectrum,
            input_table=None,
            frequency_column=self.prompt_column,
        )
        tmp_ap_file = self._get_input_data(
            input_file=afterpulse_spectrum,
            input_table=afterpulse_fitted_spectrum,
            frequency_column=self.afterpulse_column,
        )

        command = [
            f"{self.args_dict['simtel_path']}/sim_telarray/bin/norm_spe",
            "-r",
            f"{self.args_dict['step_size']},{self.args_dict['max_amplitude']}",
        ]
        if tmp_ap_file:
            command.extend(["-a", f"{tmp_ap_file.name}"])
            command.extend(["-s", f"{self.args_dict['scale_afterpulse_spectrum']}"])
            command.extend(["-t", f"{self.args_dict['afterpulse_amplitude_range'][0]}"])
        command.append(tmp_input_file.name)

        self._logger.info(f"Running norm_spe command: {' '.join(command)}")
        try:
            result = subprocess.run(command, capture_output=True, text=True, check=True)
        except subprocess.CalledProcessError as exc:
            self._logger.error(f"Error running norm_spe: {exc}")
            self._logger.error(f"stderr: {exc.stderr}")
            raise exc
        finally:
            for tmp_file in [tmp_input_file, tmp_ap_file]:
                try:
                    Path(tmp_file.name).unlink()
                except (AttributeError, FileNotFoundError):
                    pass

        self.data = result.stdout
        return result.returncode

    def _get_input_data(self, input_file, input_table, frequency_column):
        """
        Return input data in the format required by the norm_spe tool as temporary file.

        The norm_spe tool requires the data to be space separated values of the amplitude spectrum,
        with two columns: amplitude and frequency.
        Input is validated using the single_pe_spectrum schema (legacy input is not validated).

        Parameters
        ----------
        input_file : str
            Input file with amplitude spectrum.
        input_table : astro.Table
            Input table with amplitude spectrum.
        frequency_column : str
            Column name of the frequency data.
        """
        if not input_file:
            return None
        input_file = Path(input_file)

        input_data = ""
        if input_file.suffix == ".ecsv" or input_table:
            data_validator = validate_data.DataValidator(
                schema_file=self.input_schema,
                data_table=input_table,
                data_file=input_file if input_table is None else None,
            )
            table = data_validator.validate_and_transform()
            input_data = "\n".join(f"{row['amplitude']} {row[frequency_column]}" for row in table)
        else:  # legacy format
            with open(input_file, encoding="utf-8") as f:
                input_data = (
                    f.read().replace(",", " ")
                    if frequency_column == self.prompt_column
                    else f.read()
                )

        with tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8") as tmpfile:
            tmpfile.write(input_data)
        return tmpfile

    def fit_afterpulse_spectrum(self):
        """
        Fit afterpulse spectrum with a exponential decay function.

        Assume input to be in ecsv format with columns 'amplitude', 'frequency (afterpulsing)',
        and 'frequency stdev (afterpulsing)'.

        Returns
        -------
        astro.Table
            Table with fitted afterpulse spectrum data.
        """
        ap_min = self.args_dict["afterpulse_amplitude_range"][0]
        fix_k = self.args_dict.get("afterpulse_decay_factor_fixed_value")

        x, y, y_err = self._read_afterpulse_spectrum_for_fit(
            self.args_dict.get("afterpulse_spectrum"), ap_min
        )
        fit_func, p0, bounds = self.afterpulse_fit_function(fix_k=fix_k)

        result = curve_fit(fit_func, x, y, sigma=y_err, p0=p0, bounds=bounds, absolute_sigma=True)
        params, covariance = result[0], result[1]
        param_errors = np.sqrt(np.diag(covariance))
        predicted = fit_func(x, *params)
        self._afterpulse_fit_statistics(x, y, y_err, params, param_errors, predicted, fix_k)

        # table with fitted afterpulse spectrum
        x_fit = np.arange(
            ap_min, self.args_dict["afterpulse_amplitude_range"][1], self.args_dict["step_size"]
        )
        y_fit = fit_func(x_fit, *params)
        return Table([x_fit, y_fit], names=["amplitude", self.afterpulse_column])

    def afterpulse_fit_function(self, fix_k):
        """
        Afterpulse fit function: exponential decay with linear term in the exponent.

        Starting values and bounds are set for the other parameters using values typical
        for LSTN-design. Allows to fix the K parameter.

        Parameters
        ----------
        fix_K : float
            Fixed value for K parameter.

        Returns
        -------
        function
            Exponential decay function with linear term in the exponent.
        """

        def exp_decay(x, a, b, k):
            return a * np.exp(-1.0 / (b * (k / (x + k))) * x)

        p0 = [1e-5, 8.0]  # Initial guess for [A, B] typical LSTN values
        bounds_lower = [0, 0]
        bounds_upper = [1.0, 20.0]

        if fix_k is None:
            p0.append(25.0)
            bounds_lower.append(5.0)
            bounds_upper.append(35.0)
            return exp_decay, p0, (bounds_lower, bounds_upper)

        def exp_decay_fixed_k(x, a, b):
            return exp_decay(x, a, b, k=fix_k)

        return exp_decay_fixed_k, p0, (bounds_lower, bounds_upper)

    def _afterpulse_fit_statistics(self, x, y, y_err, params, param_errors, predicted, fix_k):
        """Print and return afterpulse fit statistics."""
        chi2 = np.sum(((y - predicted) / y_err) ** 2)
        ndf = len(x) - len(params)

        result = {
            "params": params.tolist(),
            "errors": param_errors.tolist(),
            "chi2_ndf": chi2 / ndf if ndf > 0 else np.nan,
        }
        if fix_k is not None:
            result["params"].append(fix_k)
            result["errors"].append(0.0)

        self._logger.info(f"Fit results: {result}")
        return result

    def _read_afterpulse_spectrum_for_fit(self, afterpulse_spectrum, fit_min_pe):
        """
        Read afterpulse spectrum data for fitting.

        Parameters
        ----------
        afterpulse_spectrum : str
            Afterpulse spectrum data file.
        fit_min_pe : float
            Minimum amplitude for fitting.

        Returns
        -------
        tuple
            Tuple with x, y, y_err data for fitting.
        """
        table = Table.read(afterpulse_spectrum, format="ascii.ecsv")
        x = table["amplitude"]
        y = table[self.afterpulse_column]
        y_err = table["frequency stdev (afterpulsing)"]
        mask = (x >= fit_min_pe) & (y > 0)
        x_fit, y_fit, y_err_fit = x[mask], y[mask], y_err[mask]
        return x_fit, y_fit, y_err_fit



#!/usr/bin/python3

r"""
Simulate pedestal events.

Use sim_telarray to simulate pedestal events for an array of telescopes.
The following types are supported:

* Pedestal events (includes night-sky background and camera noise)
* Dark pedestal events (closed camera lid, camera noise only)
* NSB-only pedestal events (open camera lid, night-sky background only, no camera noise)

Example
-------

Simulate pedestal events for Alpha North. The assumed level night-sky background is 2.0 times the
nominal value. A list of stars can be provided to simulate additional contributions.

.. code-block:: console

    simtools-simulate-pedestals --run_mode=pedestals \\
        --run_number 10 --number_of_events 1000 \\
        --array_layout_name alpha --site North \\
        --model_version 6.0.0 \\
        --zenith_angle 20 --azimuth_angle 0 \\
        --nsb_scaling_factor 2.0


Command Line Arguments
----------------------
run_mode (str, required)
    Run mode, e.g. "pedestals"
run_number (int, required)
    Run number for the simulation.
number_of_events (int, required)
    Number of calibration events to simulate.
array_layout_name (str, required)
    Array layout name, e.g. "alpha".
site (str, required)
    Site name.
model_version (str, optional)
    Version of the simulation model.
nsb_scaling_factor (float, optional)
    Scaling factor for the night-sky background rate. Default is 1.0, which
    corresponds to the nominal (dark sky) NSB rate.
stars (str, optional)
    Path to a file containing a list of stars (azimuth, zenith, weighting factor
    separated by whitespace). If provided, the stars will be used to simulate
    additional contributions to the night-sky background. For details on the
    parameters, see the sim_telarray manual.
zenith_angle (float, optional)
    Zenith angle in degrees.
azimuth_angle (float, optional)
    Azimuth angle in degrees.
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.simulator import Simulator


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__), description="Simulate calibration events."
    )
    config.parser.add_argument(
        "--run_mode",
        help="Calibration run mode",
        type=str,
        required=True,
        choices=["pedestals", "dark_pedestals", "nsb_only_pedestals"],
    )
    config.parser.add_argument(
        "--number_of_events",
        help="Number of pedestal events to simulate",
        type=int,
        required=True,
    )
    config.parser.add_argument(
        "--nsb_scaling_factor",
        help=(
            "Scaling factor for the NSB rate. "
            "Default is 1.0, which corresponds to the nominal (dark sky) NSB rate."
        ),
        type=float,
        required=False,
        default=1.0,
    )
    config.parser.add_argument(
        "--stars",
        help="List of stars (azimuth, zenith, weighting factor).",
        type=str,
        default=None,
    )

    return config.initialize(
        db_config=True,
        simulation_model=["site", "layout", "telescope", "model_version"],
        simulation_configuration={
            "corsika_configuration": ["run_number", "azimuth_angle", "zenith_angle"],
            "sim_telarray_configuration": ["all"],
        },
    )


def main():
    """Simulate pedestal events."""
    app_context = startup_application(_parse)

    simulator = Simulator(
        label=app_context.args.get("label"),
        args_dict=app_context.args,
        db_config=app_context.db_config,
    )
    simulator.simulate()


if __name__ == "__main__":
    main()



"""Simulator class for managing simulations of showers and array of telescopes."""

import gzip
import logging
import re
import shutil
from collections import defaultdict
from pathlib import Path

import numpy as np
from astropy import units as u

import simtools.utils.general as gen
from simtools.corsika.corsika_config import CorsikaConfig
from simtools.io import io_handler, table_handler
from simtools.job_execution.job_manager import JobManager
from simtools.model.array_model import ArrayModel
from simtools.runners.corsika_runner import CorsikaRunner
from simtools.runners.corsika_simtel_runner import CorsikaSimtelRunner
from simtools.simtel.simtel_io_event_writer import SimtelIOEventDataWriter
from simtools.simtel.simulator_array import SimulatorArray
from simtools.testing.sim_telarray_metadata import assert_sim_telarray_metadata
from simtools.version import semver_to_int


class InvalidRunsToSimulateError(Exception):
    """Exception for invalid runs to simulate."""


class Simulator:
    """
    Simulator is managing the simulation of showers and of the array of telescopes.

    It interfaces with simulation software packages (e.g., CORSIKA or sim_telarray).

    The configuration is set as a dict corresponding to the command line configuration groups
    (especially simulation_software, simulation_model, simulation_parameters).

    Parameters
    ----------
    args_dict : dict
        Configuration dictionary
        (includes simulation_software, simulation_model, simulation_parameters groups).
    label: str
        Instance label.
    extra_commands: str or list of str
        Extra commands to be added to the run script before the run command.
    db_config: dict
        Database configuration.
    """

    def __init__(
        self,
        args_dict,
        label=None,
        extra_commands=None,
        db_config=None,
    ):
        """Initialize Simulator class."""
        self.logger = logging.getLogger(__name__)
        self.label = label

        self.args_dict = args_dict
        self.db_config = db_config

        self.simulation_software = self.args_dict.get("simulation_software", "corsika_sim_telarray")
        self.logger.debug(f"Init Simulator {self.simulation_software}")
        self.run_mode = args_dict.get("run_mode", None)

        self.io_handler = io_handler.IOHandler()

        self.runs = self._initialize_run_list()
        self._results = defaultdict(list)
        self._test = self.args_dict.get("test", False)
        self._extra_commands = extra_commands

        self.sim_telarray_seeds = {
            "seed": self.args_dict.get("sim_telarray_instrument_seeds"),
            "random_instrument_instances": self.args_dict.get(
                "sim_telarray_random_instrument_instances"
            ),
            "seed_file_name": "sim_telarray_instrument_seeds.txt",  # name only; no directory
        }
        self.array_models = self._initialize_array_models()
        self._simulation_runner = self._initialize_simulation_runner()

    @property
    def simulation_software(self):
        """The attribute simulation_software."""
        return self._simulation_software

    @simulation_software.setter
    def simulation_software(self, simulation_software):
        """
        Set and test simulation_software type.

        Parameters
        ----------
        simulation_software: choices: [sim_telarray, corsika, corsika_sim_telarray]
            implemented are sim_telarray and CORSIKA or corsika_sim_telarray
            (running CORSIKA and piping it directly to sim_telarray)

        Raises
        ------
        ValueError

        """
        if simulation_software not in ["sim_telarray", "corsika", "corsika_sim_telarray"]:
            raise ValueError(f"Invalid simulation software: {simulation_software}")
        self._simulation_software = simulation_software.lower()

    def _initialize_array_models(self):
        """
        Initialize array simulation models.

        Returns
        -------
        list
            List of ArrayModel objects.
        """
        versions = gen.ensure_iterable(self.args_dict.get("model_version", []))

        return [
            ArrayModel(
                label=self.label,
                site=self.args_dict.get("site"),
                layout_name=self.args_dict.get("array_layout_name"),
                db_config=self.db_config,
                model_version=version,
                sim_telarray_seeds={
                    "seed": self._get_seed_for_random_instrument_instances(
                        self.sim_telarray_seeds["seed"], version
                    ),
                    "random_instrument_instances": self.sim_telarray_seeds[
                        "random_instrument_instances"
                    ],
                    "seed_file_name": self.sim_telarray_seeds["seed_file_name"],
                },
                simtel_path=self.args_dict.get("simtel_path", None),
                calibration_device_types=self._get_calibration_device_types(
                    self.args_dict.get("run_mode")
                ),
                overwrite_model_parameters=self.args_dict.get("overwrite_model_parameters", None),
            )
            for version in versions
        ]

    def _get_seed_for_random_instrument_instances(self, seed, model_version):
        """
        Generate seed for random instances of the instrument.

        Parameters
        ----------
        seed : str
            Seed string given through configuration.
        model_version: str
            Model version.

        Returns
        -------
        int
            Seed for random instances of the instrument.
        """
        if seed:
            return int(seed.split(",")[0].strip())

        seed = semver_to_int(model_version) * 10000000
        seed = seed + 1000000 if self.args_dict.get("site") != "North" else seed + 2000000
        seed = seed + (int)(self.args_dict.get("zenith_angle", 0.0 * u.deg).value) * 1000
        return seed + (int)(self.args_dict.get("azimuth_angle", 0.0 * u.deg).value)

    def _initialize_run_list(self):
        """
        Initialize run list using the configuration values.

        Uses 'run_number', 'run_number_offset' and 'number_of_runs' arguments
        to create a list of run numbers.

        Returns
        -------
        list
            List of run numbers.
        """
        offset_run_number = self.args_dict.get("run_number_offset", 0) + self.args_dict.get(
            "run_number", 1
        )
        if self.args_dict.get("number_of_runs", 1) <= 1:
            return self._prepare_run_list_and_range(
                run_list=offset_run_number,
                run_range=None,
            )
        return self._prepare_run_list_and_range(
            run_list=None,
            run_range=[
                offset_run_number,
                offset_run_number + self.args_dict["number_of_runs"],
            ],
        )

    def _prepare_run_list_and_range(self, run_list, run_range):
        """
        Prepare list of run numbers from a list or from a range.

        If both arguments are given, they will be merged into a single list.

        Attributes
        ----------
        run_list: list
            list of runs (integers)
        run_range:list
            min and max of range of runs to be simulated (two list entries)

        Returns
        -------
        list
            list of unique run numbers (integers)
        """
        if run_list is None and run_range is None:
            self.logger.debug("Nothing to prepare - run_list and run_range not given.")
            return None

        validated_runs = []
        if run_list is not None:
            validated_runs = gen.ensure_iterable(run_list)
            if not all(isinstance(r, int) for r in validated_runs):
                raise InvalidRunsToSimulateError(f"Run list must contain only integers: {run_list}")

        if run_range is not None:
            if not all(isinstance(r, int) for r in run_range) or len(run_range) != 2:
                raise InvalidRunsToSimulateError(
                    f"Run_range must contain two integers only: {run_range}"
                )
            run_range = np.arange(run_range[0], run_range[1])
            validated_runs.extend(list(run_range))

        validated_runs_unique = sorted(set(validated_runs))
        self.logger.info(f"Runlist: {validated_runs_unique}")
        return validated_runs_unique

    def _corsika_configuration(self):
        """
        Define CORSIKA configurations based on the simulation model.

        For 'corsika_sim_telarray', this is a list since multiple configurations
        might be defined to run in a single job using multipipe.

        Returns
        -------
        CorsikaConfig or list of CorsikaConfig
            CORSIKA configuration(s) based on the simulation model.
        """
        corsika_configurations = []
        for array_model in self.array_models:
            corsika_configurations.append(
                CorsikaConfig(
                    array_model=array_model,
                    label=self.label,
                    args_dict=self.args_dict,
                    db_config=self.db_config,
                    dummy_simulations=self._is_calibration_run(self.run_mode),
                )
            )
        return (
            corsika_configurations
            if self.simulation_software == "corsika_sim_telarray"
            else corsika_configurations[0]
        )

    def _initialize_simulation_runner(self):
        """
        Initialize corsika configuration and simulation runners.

        Returns
        -------
        CorsikaRunner or SimulatorArray or CorsikaSimtelRunner
            Simulation runner object.
        """
        corsika_configurations = self._corsika_configuration()

        runner_class = {
            "corsika": CorsikaRunner,
            "sim_telarray": SimulatorArray,
            "corsika_sim_telarray": CorsikaSimtelRunner,
        }.get(self.simulation_software)

        runner_args = {
            "label": self.label,
            "corsika_config": corsika_configurations,
            "simtel_path": self.args_dict.get("simtel_path"),
            "use_multipipe": runner_class is CorsikaSimtelRunner,
        }

        if runner_class is not SimulatorArray:
            runner_args["keep_seeds"] = self.args_dict.get("corsika_test_seeds", False)
        if runner_class is not CorsikaRunner:
            runner_args["sim_telarray_seeds"] = self.sim_telarray_seeds
        if runner_class is CorsikaSimtelRunner:
            runner_args["sequential"] = self.args_dict.get("sequential", False)
            runner_args["calibration_config"] = (
                self.args_dict if self._is_calibration_run(self.run_mode) else None
            )

        return runner_class(**runner_args)

    def _fill_results_without_run(self, input_file_list):
        """
        Fill results dict without calling submit (e.g., for testing).

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.
        """
        for file in gen.ensure_iterable(input_file_list):
            run = self._guess_run_from_file(file)
            self._fill_results(file, run)
            if run not in self.runs:
                self.runs.append(run)

    def simulate(self, input_file_list=None):
        """
        Submit a run script as a job.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.
        """
        runs_and_files_to_submit = self._get_runs_and_files_to_submit(
            input_file_list=input_file_list
        )
        self.logger.info(
            f"Starting submission for {len(runs_and_files_to_submit)} "
            f"run{'s' if len(runs_and_files_to_submit) > 1 else ''}"
        )

        for run_number, input_file in runs_and_files_to_submit.items():
            run_script = self._simulation_runner.prepare_run_script(
                run_number=run_number, input_file=input_file, extra_commands=self._extra_commands
            )

            job_manager = JobManager(test=self._test)
            job_manager.submit(
                run_script=run_script,
                run_out_file=self._simulation_runner.get_file_name(
                    file_type="sub_log", run_number=run_number
                ),
                log_file=self._simulation_runner.get_file_name(
                    file_type=("log"), run_number=run_number
                ),
            )

            self._fill_results(input_file, run_number)

    def _get_runs_and_files_to_submit(self, input_file_list=None):
        """
        Return a dictionary with run numbers and simulation files.

        The latter are expected to be given for the sim_telarray simulator.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        Returns
        -------
        runs_and_files: dict
            dictionary with run number as key and (if available) simulation
            file name as value

        Raises
        ------
        ValueError
            If no runs are to be submitted.
        """
        _runs_and_files = {}
        self.logger.debug(f"Getting runs and files to submit ({input_file_list})")

        if self.simulation_software == "sim_telarray" and self.run_mode is None:
            input_file_list = gen.ensure_iterable(input_file_list)
            _runs_and_files = {self._guess_run_from_file(file): file for file in input_file_list}
        else:
            _runs_and_files = dict.fromkeys(self._get_runs_to_simulate())
        if len(_runs_and_files) == 0:
            raise ValueError("No runs to submit.")
        return _runs_and_files

    def _guess_run_from_file(self, file):
        """
        Extract the run number from the given file name.

        Input file names can follow any pattern with the
        string 'run' followed by the run number.

        Parameters
        ----------
        file: Path
            Simulation file name

        Returns
        -------
        int
            The extracted run number. If extraction fails, returns 1 and logs a warning.
        """
        file_name = str(Path(file).name)

        try:
            run_str = re.search(r"run\d*", file_name).group()
            return int(run_str[3:])
        except (ValueError, AttributeError):
            self.logger.warning(f"Run number could not be guessed from {file_name} using run = 1")
            return 1

    def _fill_results(self, file, run_number):
        """
        Fill the results dict with input, output, hist, and log files.

        Parameters
        ----------
        file: str
            input file name
        run_number: int
            run number

        """
        keys = ["simtel_output", "sub_out", "log", "input", "hist", "corsika_log", "event_data"]
        results = {key: [] for key in keys}

        def get_file_name(name, **kwargs):
            return str(self._simulation_runner.get_file_name(file_type=name, **kwargs))

        if "sim_telarray" in self.simulation_software:
            results["input"].append(str(file))

        results["sub_out"].append(get_file_name("sub_log", mode="out", run_number=run_number))

        for i in range(len(self.array_models)):
            results["simtel_output"].append(
                get_file_name("simtel_output", run_number=run_number, model_version_index=i)
            )

            if "sim_telarray" in self.simulation_software:
                results["log"].append(
                    get_file_name(
                        "log",
                        simulation_software="sim_telarray",
                        run_number=run_number,
                        model_version_index=i,
                    )
                )
                results["hist"].append(
                    get_file_name(
                        "histogram",
                        simulation_software="sim_telarray",
                        run_number=run_number,
                        model_version_index=i,
                    )
                )
                results["event_data"].append(
                    get_file_name(
                        "event_data",
                        simulation_software="sim_telarray",
                        run_number=run_number,
                        model_version_index=i,
                    )
                )

            if "corsika" in self.simulation_software:
                results["corsika_log"].append(
                    get_file_name(
                        "corsika_log",
                        simulation_software="corsika",
                        run_number=run_number,
                        model_version_index=i,
                    )
                )

        for key in keys:
            self._results[key].extend(results[key])

    def get_file_list(self, file_type="simtel_output"):
        """
        Get list of files generated by simulations.

        Options are "input", "simtel_output", "hist", "log", "corsika_log".
        Not all file types are available for all simulation types.
        Returns an empty list for an unknown file type.

        Parameters
        ----------
        file_type : str
            File type to be listed.

        Returns
        -------
        list
            List with the full path of all output files.

        """
        self.logger.info(f"Getting list of {file_type} files")
        return self._results[file_type]

    def _make_resources_report(self, input_file_list):
        """
        Prepare a simple report on computing wall clock time used in the simulations.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        Returns
        -------
        dict
           Dictionary with reports on computing resources

        """
        if len(self._results["sub_out"]) == 0:
            if input_file_list is None:
                return {"Wall time/run [sec]": np.nan}
            self._fill_results_without_run(input_file_list)

        runtime = []

        _resources = {}
        for run in self.runs:
            _resources = self._simulation_runner.get_resources(run_number=run)
            if _resources.get("runtime"):
                runtime.append(_resources["runtime"])

        mean_runtime = np.mean(runtime)

        resource_summary = {}
        resource_summary["Wall time/run [sec]"] = mean_runtime
        if "n_events" in _resources and _resources["n_events"] > 0:
            resource_summary["#events/run"] = _resources["n_events"]
            resource_summary["Wall time/1000 events [sec]"] = (
                mean_runtime * 1000 / _resources["n_events"]
            )

        return resource_summary

    def resources(self, input_file_list=None):
        """
        Print a simple report on computing resources used.

        Includes run time per run only at this point.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        """
        resources = self._make_resources_report(input_file_list)
        print("-----------------------------")
        print(f"Computing Resources Report - {self.simulation_software} Simulations")
        for key, value in resources.items():
            print(f"{key} = {value:.2f}")
        print("-----------------------------")

    def _get_runs_to_simulate(self, run_list=None, run_range=None):
        """
        Process run_list and run_range and return the validated list of runs.

        Attributes
        ----------
        run_list: list
            list of runs (integers)
        run_range:list
            min and max of range of runs to be simulated (two list entries)

        Returns
        -------
        list
            list of unique run numbers (integers)

        """
        if run_list is None and run_range is None:
            return [] if self.runs is None else self.runs
        return self._prepare_run_list_and_range(run_list, run_range)

    def save_file_lists(self):
        """Save files lists for output and log files."""
        for file_type in ["simtel_output", "log", "corsika_log", "hist"]:
            file_name = self.io_handler.get_output_directory().joinpath(f"{file_type}_files.txt")
            file_list = self.get_file_list(file_type=file_type)
            if all(element is not None for element in file_list) and len(file_list) > 0:
                self.logger.info(f"Saving list of {file_type} files to {file_name}")
                with open(file_name, "w", encoding="utf-8") as f:
                    for line in self.get_file_list(file_type=file_type):
                        f.write(f"{line}\n")
            else:
                self.logger.debug(f"No files to save for {file_type} files.")

    def save_reduced_event_lists(self):
        """
        Save reduced event lists with event data on simulated and triggered events.

        The files are saved with the same name as the sim_telarray output file
        but with a 'hdf5' extension.
        """
        if "sim_telarray" not in self.simulation_software:
            self.logger.warning(
                "Reduced event lists can only be saved for sim_telarray simulations."
            )
            return

        input_files = self.get_file_list(file_type="simtel_output")
        output_files = self.get_file_list(file_type="event_data")
        for input_file, output_file in zip(input_files, output_files):
            generator = SimtelIOEventDataWriter([input_file])
            table_handler.write_tables(
                tables=generator.process_files(),
                output_file=Path(output_file),
                overwrite_existing=True,
            )

    def pack_for_register(self, directory_for_grid_upload=None):
        """
        Pack simulation output files for registering on the grid.

        Creates separate tarballs for each model version's log files.

        Parameters
        ----------
        directory_for_grid_upload: str
            Directory for the tarball with output files.

        """
        self.logger.info(
            f"Packing the output files for registering on the grid ({directory_for_grid_upload})"
        )
        output_files = self.get_file_list(file_type="simtel_output")
        log_files = self.get_file_list(file_type="log")
        corsika_log_files = self.get_file_list(file_type="corsika_log")
        histogram_files = self.get_file_list(file_type="hist")
        reduced_event_files = (
            self.get_file_list(file_type="event_data")
            if self.args_dict.get("save_reduced_event_lists")
            else []
        )

        directory_for_grid_upload = (
            Path(directory_for_grid_upload)
            if directory_for_grid_upload
            else self.io_handler.get_output_directory().joinpath("directory_for_grid_upload")
        )
        directory_for_grid_upload.mkdir(parents=True, exist_ok=True)

        # If there are more than one model version,
        # duplicate the corsika log file to have one for each model version with the "right name".
        if len(self.array_models) > 1 and corsika_log_files:
            self._copy_corsika_log_file_for_all_versions(corsika_log_files)

        # Group files by model version
        for model in self.array_models:
            model_version = model.model_version
            model_files = gen.ensure_iterable(model.pack_model_files())

            # Filter files for this model version
            model_logs = [f for f in log_files if model_version in f]
            model_hists = [f for f in histogram_files if model_version in f]
            model_corsika_logs = [f for f in corsika_log_files if model_version in f]

            if model_logs:
                tar_file_name = Path(model_logs[0]).name.replace("log.gz", "log_hist.tar.gz")
                tar_file_path = directory_for_grid_upload.joinpath(tar_file_name)
                # Add all relevant model, log, histogram, and CORSIKA log files to the tarball
                files_to_tar = model_logs + model_hists + model_corsika_logs + model_files
                gen.pack_tar_file(tar_file_path, files_to_tar)

        for file_to_move in output_files + reduced_event_files:
            source_file = Path(file_to_move)
            destination_file = directory_for_grid_upload / source_file.name
            if destination_file.exists():
                self.logger.warning(f"Overwriting existing file: {destination_file}")
            shutil.move(source_file, destination_file)

        self.logger.info(f"Output files for the grid placed in {directory_for_grid_upload!s}")

    def validate_metadata(self):
        """Validate metadata in the sim_telarray output files."""
        if "sim_telarray" not in self.simulation_software:
            self.logger.info("No sim_telarray files to validate.")
            return

        for model in self.array_models:
            files = self.get_file_list(file_type="simtel_output")
            output_file = next((f for f in files if model.model_version in f), None)
            if output_file:
                self.logger.info(f"Validating metadata for {output_file}")
                assert_sim_telarray_metadata(output_file, model)
                self.logger.info(f"Metadata for sim_telarray file {output_file} is valid.")
            else:
                self.logger.warning(
                    f"No sim_telarray file found for model version {model.model_version}: {files}"
                )

    def _copy_corsika_log_file_for_all_versions(self, corsika_log_files):
        """
        Create copies of the CORSIKA log file for each model version.

        Adds a header comment to each copy explaining its relationship to the original.

        Parameters
        ----------
        corsika_log_files: list
            List containing the original CORSIKA log file path.
        """
        original_log = Path(corsika_log_files[0])
        # Find which model version the original log belongs to
        original_version = next(
            model.model_version
            for model in self.array_models
            if re.search(
                rf"(?<![0-9A-Za-z]){re.escape(model.model_version)}(?![0-9A-Za-z])",
                original_log.name,
            )
        )

        for model in self.array_models:
            if model.model_version == original_version:
                continue

            new_log = original_log.parent / original_log.name.replace(
                original_version, model.model_version
            )

            with gzip.open(new_log, "wt", encoding="utf-8") as new_file:
                # Write the header to the new file
                header = (
                    f"###############################################################\n"
                    f"Copy of CORSIKA log file from model version {original_version}.\n"
                    f"Applicable also for {model.model_version} (same CORSIKA configuration,\n"
                    f"different sim_telarray model versions in the same run).\n"
                    f"###############################################################\n\n"
                )
                new_file.write(header)

                # Copy the content of the original log file, ignoring invalid characters
                with gzip.open(original_log, "rt", encoding="utf-8", errors="ignore") as orig_file:
                    for line in orig_file:
                        new_file.write(line)

            corsika_log_files.append(str(new_log))

    @staticmethod
    def _is_calibration_run(run_mode):
        """
        Check if this simulation is a calibration run.

        Parameters
        ----------
        run_mode: str
            Run mode of the simulation.

        Returns
        -------
        bool
            True if it is a calibration run, False otherwise.
        """
        return run_mode in [
            "pedestals",
            "dark_pedestals",
            "nsb_only_pedestals",
            "direct_injection",
        ]

    @staticmethod
    def _get_calibration_device_types(run_mode):
        """
        Get the list of calibration device types based on the run mode.

        Parameters
        ----------
        run_mode: str
            Run mode of the simulation.

        Returns
        -------
        list
            List of calibration device types.
        """
        if run_mode == "direct_injection":
            return ["flat_fielding"]
        return []



#!/usr/bin/python3

r"""
Verify simulation model production tables and model parameters for completeness.

This application is a utility to be used in the CI pipeline of the SimulationModels
repository. It checks that all model parameters defined in the production tables
exist in the simulation models repository.

Example
-------

.. code-block:: console

    simtools-maintain-simulation-model-verify-production-tables \\
        --simulation_models_path /path/to/simulation/models/repository

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model import model_repository


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Verify simulation model production tables and model parameters for completeness. "
            "This application checks that all model parameters defined in the production tables "
            "exist in the simulation models repository."
        ),
    )
    config.parser.add_argument(
        "--simulation_models_path",
        help="Path to the simulation models repository.",
        type=str,
        required=True,
    )
    return config.initialize(db_config=False, output=False, paths=False)


def main():
    """Verify simulation model production tables."""
    app_context = startup_application(_parse)

    if not model_repository.verify_simulation_model_production_tables(
        simulation_models_path=app_context.args["simulation_models_path"]
    ):
        raise RuntimeError(
            "Verification failed: Some model parameters are missing in the repository."
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Module for visualization."""

import logging
import re
from collections import OrderedDict
from pathlib import Path

import astropy.units as u
import matplotlib.pyplot as plt
from astropy.table import QTable
from cycler import cycler
from matplotlib import gridspec

COLORS = {}
COLORS["classic"] = [
    "#ba2c54",
    "#5B90DC",
    "#FFAB44",
    "#0C9FB3",
    "#57271B",
    "#3B507D",
    "#794D88",
    "#FD6989",
    "#8A978E",
    "#3B507D",
    "#D8153C",
    "#cc9214",
]
COLORS["modified classic"] = [
    "#D6088F",
    "#424D9C",
    "#178084",
    "#AF99DA",
    "#F58D46",
    "#634B5B",
    "#0C9FB3",
    "#7C438A",
    "#328cd6",
    "#8D0F25",
    "#8A978E",
    "#ffcb3d",
]
COLORS["autumn"] = [
    "#A9434D",
    "#4E615D",
    "#3C8DAB",
    "#A4657A",
    "#424D9C",
    "#DC575A",
    "#1D2D38",
    "#634B5B",
    "#56276D",
    "#577580",
    "#134663",
    "#196096",
]
COLORS["purples"] = [
    "#a57bb7",
    "#343D80",
    "#EA60BF",
    "#B7308E",
    "#E099C3",
    "#7C438A",
    "#AF99DA",
    "#4D428E",
    "#56276D",
    "#CC4B93",
    "#DC4E76",
    "#5C4AE4",
]
COLORS["greens"] = [
    "#268F92",
    "#abc14d",
    "#8A978E",
    "#0C9FB3",
    "#BDA962",
    "#B0CB9E",
    "#769168",
    "#5E93A5",
    "#178084",
    "#B7BBAD",
    "#163317",
    "#76A63F",
]

COLORS["default"] = COLORS["classic"]

MARKERS = ["o", "s", "v", "^", "*", "P", "d", "X", "p", "<", ">", "h"]
LINES = [
    (0, ()),  # solid
    (0, (1, 1)),  # densely dotted
    (0, (3, 1, 1, 1)),  # densely dashdotted
    (0, (5, 5)),  # dashed
    (0, (3, 1, 1, 1, 1, 1)),  # densely dashdotdotted
    (0, (5, 1)),  # densely dashed
    (0, (1, 5)),  # dotted
    (0, (3, 5, 1, 5)),  # dashdotted
    (0, (3, 5, 1, 5, 1, 5)),  # dashdotdotted
    (0, (5, 10)),  # loosely dashed
    (0, (1, 10)),  # loosely dotted
    (0, (3, 10, 1, 10)),  # loosely dashdotted
]

_logger = logging.getLogger(__name__)


def _add_unit(title, array):
    """
    Add a unit to "title" (presumably an axis title).

    The unit is extracted from the unit field of the array, in case array is an astropy quantity.
    If a unit is found, it is added to title in the form [unit]. If a unit already is present in
    title (in the same form), a warning is printed and no unit is added. The function assumes
    array not to be empty and returns the modified title.

    Parameters
    ----------
    title: str
    array: astropy.Quantity

    Returns
    -------
    str
        Title with units.
    """
    if title and "[" in title and "]" in title:
        _logger.warning(
            "Tried to add a unit from astropy.unit, "
            "but axis already has an explicit unit. Left axis title as is."
        )
        return title

    if not isinstance(array, u.Quantity) or not str(array[0].unit):
        return title

    unit = str(array[0].unit)
    unit_str = f" [{unit}]"
    if re.search(r"\d", unit_str):
        unit_str = re.sub(r"(\d)", r"^\1", unit_str)
        unit_str = unit_str.replace("[", r"[$").replace("]", r"$]")

    return f"{title}{unit_str}"


def set_style(palette="default", big_plot=False):
    """
    Set the plotting style to homogenize plot style across the framework.

    The function receives the colour palette name and whether it is a big plot or not.\
    The latter sets the fonts and marker to be bigger in case it is a big plot. The available \
    colour palettes are as follows:

    - classic (default): A classic colorful palette with strong colors and contrast.
    - modified classic: Similar to the classic, with slightly different colors.
    - autumn: A slightly darker autumn style colour palette.
    - purples: A pseudo sequential purple colour palette (not great for contrast).
    - greens: A pseudo sequential green colour palette (not great for contrast).

    To use the function, simply call it before plotting anything. The function is made public, so \
    that it can be used outside the visualize module. However, it is highly recommended to create\
    plots only through the visualize module.

    Parameters
    ----------
    palette: str
        Colour palette.
    big_plot: bool
        Flag to set fonts and marker bigger. If True, it sets them bigger.

    Raises
    ------
    KeyError
        if provided palette does not exist.
    """
    if palette not in COLORS:
        raise KeyError(f"palette must be one of {', '.join(COLORS)}")

    fontsize = {"default": 17, "big_plot": 30}
    markersize = {"default": 6, "big_plot": 18}
    plot_size = "big_plot" if big_plot else "default"

    plt.rc("lines", linewidth=2, markersize=markersize[plot_size])
    plt.rc(
        "axes",
        prop_cycle=(
            cycler(color=COLORS[palette]) + cycler(linestyle=LINES) + cycler(marker=MARKERS)
        ),
        titlesize=fontsize[plot_size],
        labelsize=fontsize[plot_size],
        labelpad=5,
        grid=True,
        axisbelow=True,
    )
    plt.rc("xtick", labelsize=fontsize[plot_size])
    plt.rc("ytick", labelsize=fontsize[plot_size])
    plt.rc("legend", loc="best", shadow=False, fontsize="medium")
    plt.rc("font", family="serif", size=fontsize[plot_size])


def get_colors(palette="default"):
    """
    Get the colour list of the palette requested.

    If no palette is provided, the default is returned.

    Parameters
    ----------
    palette: str
        Colour palette.

    Returns
    -------
    list
        Colour list.

    Raises
    ------
    KeyError
        if provided palette does not exist.
    """
    if palette not in COLORS:
        raise KeyError(f"palette must be one of {', '.join(COLORS)}")

    return COLORS[palette]


def get_markers():
    """
    Get the marker list used in this module.

    Returns
    -------
    list
        List with markers.
    """
    return MARKERS


def get_lines():
    """
    Get the line style list used in this module.

    Returns
    -------
    list
        List with line styles.
    """
    return LINES


def filter_plot_kwargs(kwargs):
    """Filter out kwargs that are not valid for plt.plot."""
    valid_keys = {
        "color",
        "linestyle",
        "linewidth",
        "marker",
        "markersize",
        "markerfacecolor",
        "markeredgecolor",
        "markeredgewidth",
        "alpha",
        "label",
        "zorder",
        "dashes",
        "gapcolor",
        "solid_capstyle",
        "solid_joinstyle",
        "dash_capstyle",
        "dash_joinstyle",
        "antialiased",
    }
    return {k: v for k, v in kwargs.items() if k in valid_keys}


def plot_1d(data, **kwargs):
    """
    Produce a high contrast one dimensional plot from multiple data sets.

    A ratio plot can be added at the bottom to allow easy comparison. Additional options,
    such as plot title, plot legend, etc., are given in kwargs. Any option that can be
    changed after plotting (e.g., axes limits, log scale, etc.) should be done using the
    returned plt instance.

    Parameters
    ----------
    data: numpy structured array or a dictionary of structured arrays
          Each structured array has two columns, the first is the x-axis and the second the y-axis.
          The titles of the columns are set as the axes titles.
          The labels of each dataset set are given in the dictionary keys
          and will be used in the legend.
    **kwargs:
        * palette: string
          Choose a colour palette (see set_style for additional information).
        * title: string
          Set a plot title.
        * no_legend: bool
          Do not print a legend for the plot.
        * big_plot: bool
          Increase marker and font sizes (like in a wide light curve).
        * no_markers: bool
          Do not print markers.
        * empty_markers: bool
          Print empty (hollow) markers
        * plot_ratio: bool
          Add a ratio plot at the bottom. The first entry in the data dictionary
          is used as the reference for the ratio.
          If data dictionary is not an OrderedDict, the reference will be random.
        * plot_difference: bool
          Add a difference plot at the bottom. The first entry in the data dictionary
          is used as the reference for the difference.
          If data dictionary is not an OrderedDict, the reference will be random.
        * Any additional kwargs for plt.plot

    Returns
    -------
    pyplot.figure
        Instance of pyplot.figure in which the plot was produced

    Raises
    ------
    ValueError
        if asked to plot a ratio or difference with just one set of data
    """
    kwargs = handle_kwargs(kwargs)
    data_dict, plot_ratio, plot_difference = prepare_data(data, kwargs)
    fig, ax1, gs = setup_plot(kwargs, plot_ratio, plot_difference)
    plot_args = filter_plot_kwargs(kwargs)
    plot_main_data(data_dict, kwargs, plot_args)

    if plot_ratio or plot_difference:
        plot_ratio_difference(ax1, data_dict, plot_ratio, gs, plot_args)

    if not (plot_ratio or plot_difference):
        plt.tight_layout()

    return fig


def handle_kwargs(kwargs):
    """Extract and handle keyword arguments."""
    kwargs_defaults = {
        "palette": "default",
        "big_plot": False,
        "title": "",
        "no_legend": False,
        "no_markers": False,
        "empty_markers": False,
        "plot_ratio": False,
        "plot_difference": False,
        "xscale": "linear",
        "yscale": "linear",
        "xlim": (None, None),
        "ylim": (None, None),
        "xtitle": None,
        "ytitle": None,
    }
    for key, default in kwargs_defaults.items():
        kwargs[key] = kwargs.pop(key, default)

    if kwargs["no_markers"]:
        kwargs["marker"] = "None"
        kwargs["linewidth"] = 4
    if kwargs["empty_markers"]:
        kwargs["markerfacecolor"] = "None"

    return kwargs


def prepare_data(data, kwargs):
    """Prepare data for plotting."""
    if not isinstance(data, dict):
        data_dict = {"_default": data}
    else:
        data_dict = data

    if (kwargs["plot_ratio"] or kwargs["plot_difference"]) and len(data_dict) < 2:
        raise ValueError("Asked to plot a ratio or difference with just one set of data")

    return data_dict, kwargs["plot_ratio"], kwargs["plot_difference"]


def setup_plot(kwargs, plot_ratio, plot_difference):
    """Set up the plot style, figure, and gridspec."""
    set_style(kwargs["palette"], kwargs["big_plot"])

    if plot_ratio or plot_difference:
        gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
        fig = plt.figure(figsize=(8, 8))
        gs.update(hspace=0.02)
    else:
        gs = gridspec.GridSpec(1, 1)
        fig = plt.figure(figsize=(8, 6))

    plt.subplot(gs[0])
    ax1 = plt.gca()

    return fig, ax1, gs


def _plot_error_plots(kwargs, data_now, x_col, y_col, x_err_col, y_err_col, color):
    """Plot error plots."""
    if kwargs.get("error_type") == "fill_between" and y_err_col:
        plt.fill_between(
            data_now[x_col],
            data_now[y_col] - data_now[y_err_col],
            data_now[y_col] + data_now[y_err_col],
            alpha=0.2,
            color=color,
        )

    if kwargs.get("error_type") == "errorbar" and (x_err_col or y_err_col):
        plt.errorbar(
            data_now[x_col],
            data_now[y_col],
            xerr=data_now[x_err_col] if x_err_col else None,
            yerr=data_now[y_err_col] if y_err_col else None,
            fmt=".",
            color=color,
        )


def _get_data_columns(data_now):
    """Return data columns depending on availability."""
    columns = data_now.dtype.names
    assert len(columns) >= 2, "Input array must have at least two columns with titles."
    x_col, y_col = columns[:2]
    if len(columns) == 3:
        x_err_col = None
        y_err_col = columns[2]
    elif len(columns) == 4:
        x_err_col = columns[2]
        y_err_col = columns[3]
    else:
        x_err_col = None
        y_err_col = None
    return x_col, y_col, x_err_col, y_err_col


def plot_main_data(data_dict, kwargs, plot_args):
    """Plot the main data."""
    for label, data_now in data_dict.items():
        x_col, y_col, x_err_col, y_err_col = _get_data_columns(data_now)

        x_title = kwargs.get("xtitle", x_col)
        y_title = kwargs.get("ytitle", y_col)

        x_title_unit = _add_unit(x_title, data_now[x_col])
        y_title_unit = _add_unit(y_title, data_now[y_col])

        (line,) = plt.plot(data_now[x_col], data_now[y_col], label=label, **plot_args)
        color = line.get_color()

        _plot_error_plots(kwargs, data_now, x_col, y_col, x_err_col, y_err_col, color)

    plt.xscale(kwargs["xscale"])
    plt.yscale(kwargs["yscale"])
    plt.xlim(kwargs["xlim"])
    plt.ylim(kwargs["ylim"])
    plt.ylabel(y_title_unit)

    if not (kwargs["plot_ratio"] or kwargs["plot_difference"]):
        plt.xlabel(x_title_unit)

    if kwargs["title"]:
        plt.title(kwargs["title"], y=1.02)

    if "_default" not in data_dict and not kwargs["no_legend"]:
        plt.legend()


def plot_ratio_difference(ax1, data_dict, plot_ratio, gs, plot_args):
    """Plot the ratio or difference plot."""
    plt.subplot(gs[1], sharex=ax1)
    plt.plot([], [])  # Advance cycler for consistent colors/styles

    data_ref_name = next(iter(data_dict))
    for label, data_now in data_dict.items():
        if label == data_ref_name:
            continue
        x_title, y_title = data_now.dtype.names[0], data_now.dtype.names[1]
        x_title_unit = _add_unit(x_title, data_now[x_title])

        if plot_ratio:
            y_values = data_now[y_title] / data_dict[data_ref_name][y_title]
            y_title_ratio = f"Ratio to {data_ref_name}"
        else:
            y_values = data_now[y_title] - data_dict[data_ref_name][y_title]
            y_title_ratio = f"Difference to {data_ref_name}"

        plt.plot(data_now[x_title], y_values, **plot_args)

    plt.xlabel(x_title_unit)

    if len(y_title_ratio) > 20 and plot_ratio:
        y_title_ratio = "Ratio"

    plt.ylabel(y_title_ratio)
    ylim = plt.gca().get_ylim()
    nbins = min(int((ylim[1] - ylim[0]) / 0.05 + 1), 6)
    plt.locator_params(axis="y", nbins=nbins)


def plot_table(table, y_title, **kwargs):
    """
    Produce a high contrast one dimensional plot from the data in an astropy.Table.

    A ratio plot can be added at the bottom to allow easy comparison. Additional options, such
    as plot title, plot legend, etc., are given in kwargs. Any option that can be changed after
    plotting (e.g., axes limits, log scale, etc.) should be done using the returned plt instance.

    Parameters
    ----------
    table: astropy.Table or astropy.QTable
           The first column of the table is the x-axis and the second column is the y-axis. Any \
           additional columns will be treated as additional data to plot. The column titles are \
           used in the legend (except for the first column).
    y_title: str
           The y-axis title.

    **kwargs:
        * palette: choose a colour palette (see set_style for additional information).
        * title: set a plot title.
        * no_legend: do not print a legend for the plot.
        * big_plot: increase marker and font sizes (like in a wide light curve).
        * no_markers: do not print markers.
        * empty_markers: print empty (hollow) markers
        * plot_ratio: bool
          Add a ratio plot at the bottom. The first entry in the data dictionary is used as the \
          reference for the ratio. If data dictionary is not an OrderedDict, the reference will be\
          random.
        * plot_difference: bool
          Add a difference plot at the bottom. The first entry in the data dictionary is used as \
          the reference for the difference. If data dictionary is not an OrderedDict, the reference\
          will be random.
        * Any additional kwargs for plt.plot

    Returns
    -------
    pyplot.fig
        Instance of pyplot.fig.

    Raises
    ------
    ValueError
        if table has less than two columns.
    """
    if len(table.keys()) < 2:
        raise ValueError("Table has to have at least two columns")

    x_axis = table.keys()[0]
    data_dict = OrderedDict()
    for column in table.keys()[1:]:
        data_dict[column] = QTable([table[x_axis], table[column]], names=[x_axis, y_title])

    return plot_1d(data_dict, **kwargs)


def plot_hist_2d(data, **kwargs):
    """
    Produce a two dimensional histogram plot.

    Any option that can be changed after plotting (e.g., axes limits, log scale, etc.)
    should be done using the returned plt instance.

    Parameters
    ----------
    data: numpy structured array
          The columns of the structured array are used as the x-axis and y-axis titles.
    **kwargs:
        * title: set a plot title.
        * Any additional kwargs for plt.hist2d

    Returns
    -------
    pyplot.figure
        Instance of pyplot.figure in which the plot was produced.

    """
    cmap = plt.cm.gist_heat_r
    if "title" in kwargs:
        title = kwargs["title"]
        kwargs.pop("title", None)
    else:
        title = ""

    # Set default style since the usual options do not affect 2d plots (for now).
    set_style()

    gs = gridspec.GridSpec(1, 1)
    fig = plt.figure(figsize=(8, 6))

    ##########################################################################################
    # Plot the data
    ##########################################################################################

    plt.subplot(gs[0])
    assert len(data.dtype.names) == 2, "Input array must have two columns with titles."
    x_title, y_title = data.dtype.names[0], data.dtype.names[1]
    x_title_unit = _add_unit(x_title, data[x_title])
    y_title_unit = _add_unit(y_title, data[y_title])
    plt.hist2d(data[x_title], data[y_title], cmap=cmap, **kwargs)

    plt.xlabel(x_title_unit)
    plt.ylabel(y_title_unit)

    plt.gca().set_aspect("equal", adjustable="datalim")

    if len(title) > 0:
        plt.title(title, y=1.02)

    fig.tight_layout()

    return fig


def save_figure(fig, output_file, figure_format=None, log_title="", dpi="figure"):
    """
    Save figure to output file(s).

    Parameters
    ----------
    fig: plt.figure
        Figure to save.
    output_file: Path, str
        Path to save the figure (without suffix).
    figure_format: list
        List of formats to save the figure.
    title: str
        Title of the figure to be added to the log message.
    """
    figure_format = figure_format or ["pdf", "png"]
    for fmt in figure_format:
        _file = Path(output_file).with_suffix(f".{fmt}")
        fig.savefig(_file, format=fmt, bbox_inches="tight", dpi=dpi)
        logging.info(f"Saved plot {log_title} to {_file}")

    fig.clf()
